"""
Tests for Docking
"""
import os
import platform
import unittest
import pytest
import logging
import numpy as np
import jaqpotpy as jp
from jaqpotpy.descriptors import ComplexFeaturizer
from jaqpotpy.models import Model
from jaqpotpy.docking.pose_generation import PoseGenerator
from jaqpotpy.docking.utils import create_hydrated_pdbqt_pdb

IS_WINDOWS = platform.system() == 'Windows'


class TestDocking(unittest.TestCase):
  """
  Does sanity checks on pose generation.
  """

  def setUp(self):
    current_dir = os.path.dirname(os.path.realpath(__file__))
    # self.protein_file = os.path.join(current_dir, "1jld_protein.pdb")
    # self.protein_file = os.path.join(current_dir, "7zb6.pdb")
    # self.protein_file = os.path.join(current_dir, "0_a.pdb")
    self.protein_file = os.path.join(current_dir, "117_a.pdb")
    self.ligand_file = os.path.join(current_dir, "mulno.sdf")
    # self.ligand_file = os.path.join(current_dir, "1jld_ligand.sdf")

    # self.protein_file = os.path.join(current_dir, "1a9m_pocket.pdb")
    # self.ligand_file = os.path.join(current_dir, "1a9m_ligand.sdf")
    # self.ligand_file = os.path.join(current_dir, "ZINC000787318646.sdf")

  @pytest.mark.slow
  def test_docker_init(self):
    """Test that Docker can be initialized."""
    vpg = jp.docking.VinaPoseGenerator()
    jp.docking.Docker(vpg)

  def test_tdc(self):
    from jaqpotpy.datasets import TorchGraphDataset
    from jaqpotpy.models import Evaluator
    import torch
    from jaqpotpy.models import GCN_V1, MolecularTorchGeometric, AttentiveFP

    from jaqpotpy.descriptors.molecular import MolGraphConvFeaturizer, PagtnMolGraphFeaturizer, AttentiveFPFeaturizer

    from sklearn.metrics import accuracy_score, f1_score, roc_auc_score

    from tdc.single_pred import ADME
    data = ADME(name='BBB_Martins')
    split = data.get_split()
    featurizer = AttentiveFPFeaturizer()
    train_dataset = TorchGraphDataset(smiles=split['train']['Drug'], y=split['train']['Y'], task='classification'
                                , featurizer=featurizer)

    test_dataset = TorchGraphDataset(smiles=split['test']['Drug'], y=split['test']['Y'], task='classification'
                                , featurizer=featurizer)

    train_dataset.create()
    test_dataset.create()
    val = Evaluator()
    val.dataset = test_dataset
    val.register_scoring_function('Accuracy', accuracy_score)
    val.register_scoring_function('F1', f1_score)
    val.register_scoring_function('Roc Auc', roc_auc_score)

    model = AttentiveFP(in_channels=39, hidden_channels=100, out_channels=2, edge_dim=10, num_layers=6,
                        num_timesteps=2).jittable()

    optimizer = torch.optim.Adam(model.parameters(), lr=0.0001, weight_decay=5e-4)
    criterion = torch.nn.CrossEntropyLoss()
    m = MolecularTorchGeometric(dataset=train_dataset
                                , model_nn=model, eval=val
                                , train_batch=262, test_batch=200
                                , epochs=840, optimizer=optimizer, criterion=criterion, device="cpu").fit()

    m.eval()

    mol_m = m.create_molecular_model()
    mol_m.model_title = "BBB_MODEL"
    mol_m.save()

  def test_smi_file(self):
      from rdkit import Chem
      from rdkit.Chem import AllChem
      suppl = Chem.SmilesMolSupplier("/Users/pantelispanka/Downloads/bbb_v01.smi")
      for mol in suppl:
          id = mol.GetProp('id')
          smile = Chem.MolToSmiles(mol)
          mol = Chem.AddHs(mol)
          AllChem.EmbedMolecule(mol, randomSeed=0xf00d)
          Chem.MolToMolBlock(mol)
          # ligand = "./" + str(id) + "_a_syn_ligand.sdf"
          ligand = "/Users/pantelispanka/Jaqpot/jaqpotpy/jaqpotpy/docking/test/bbb_v01_0_a_syn_ligand.sdf"
          # with Chem.SDWriter(ligand) as w:
          #     w.write(mol)
          vpg = jp.docking.VinaPoseGenerator(prepare_protein=True)
          docker = jp.docking.Docker(vpg)
          docked_outputs = docker.dock((self.protein_file, ligand),
                                       centroid=(80, 38, 48),
                                       box_dims=(12, 12, 12),
                                       exhaustiveness=12,
                                       num_modes=1,
                                       out_dir="./tmp",
                                       use_pose_generator_scores=True)

          # Check only one output since num_modes==1
          docked_outputs = list(docked_outputs)
          print(docked_outputs)
          print(len(docked_outputs))
          print(len(docked_outputs[0]))
          assert len(docked_outputs) == 1
          assert len(docked_outputs[0]) == 2


  def test_jaqpotpy(self):
    import warnings
    warnings.filterwarnings("ignore")
    import pandas as pd
    from rdkit import Chem
    from jaqpotpy import Jaqpot
    from jaqpotpy.models import MolecularModel
    jaqpot = Jaqpot()
    jaqpot.request_key("pantelispanka@gmail.com", "kapan2")
    model = MolecularModel().load_from_jaqpot(jaqpot, "oZZfU6RQgLnmHgk88hnc")
    model.save()
    smiles_file = "/Users/pantelispanka/guacamol_v1_all.csv"
    df = pd.read_csv(smiles_file)
    for index, smile in enumerate(df[:100000]['SMILES']):
        print(index)
        print(smile)
        mol = Chem.MolFromSmiles(smile)
        try:
            model(mol)
            print(model.prediction)
            print(model.probability)
        except Exception as e:
            print(str(e))

  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_docker_dock(self):
    """Test that Docker can dock."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator(calc_charges=False, add_hydrogens=False)
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=12,
                                 num_modes=4,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    print(docked_outputs)
    print(list(docked_outputs))
    # Check only one output since num_modes==1
    assert len(list(docked_outputs)) == 4

  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_docker_pose_generator_scores_with_smiles(self):
    from rdkit import Chem
    from rdkit.Chem import AllChem
    import pandas as pd
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score

    smiles_file = "/Users/pantelispanka/guacamol_v1_all.csv"
    df = pd.read_csv(smiles_file)
    for index, smile in enumerate(df[:100]['SMILES']):
        print(index)
        print(smile)
        # pdb_qt = create_hydrated_pdbqt_pdb(self.protein_file)
        # smile = "C1CNCCN(C1)S(=O)(=O)C2=CC=CC3=C2C=CN=C3"
        mol = Chem.MolFromSmiles(smile)
        mol = Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol, randomSeed=0xf00d)
        Chem.MolToMolBlock(mol)
        ligand = "./" + str(index) + "_a_syn_ligand.sdf"
        with Chem.SDWriter(ligand) as w:
            w.write(mol)
        vpg = jp.docking.VinaPoseGenerator(prepare_protein=True)
        docker = jp.docking.Docker(vpg)
        docked_outputs = docker.dock((self.protein_file, ligand),
                                     centroid=(80, 38, 48),
                                     box_dims=(12, 12, 12),
                                     exhaustiveness=12,
                                     num_modes=1,
                                     out_dir="./tmp",
                                     use_pose_generator_scores=True)

        # Check only one output since num_modes==1
        docked_outputs = list(docked_outputs)
        print(docked_outputs)
        print(len(docked_outputs))
        print(len(docked_outputs[0]))
        assert len(docked_outputs) == 1
        assert len(docked_outputs[0]) == 2


  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_docker_pose_generator_scores_with_pdbqt(self):
    from rdkit import Chem
    from rdkit.Chem import AllChem
    import pandas as pd
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score

    # pdb_qt = create_hydrated_pdbqt_pdb(self.protein_file)
    smile = "C1CNCCN(C1)S(=O)(=O)C2=CC=CC3=C2C=CN=C3"
    mol = Chem.MolFromSmiles(smile)
    mol = Chem.AddHs(mol)
    AllChem.EmbedMolecule(mol, randomSeed=0xf00d)
    Chem.MolToMolBlock(mol)
    ligand = "./fasudil.sdf"
    with Chem.SDWriter(ligand) as w:
        w.write(mol)
    vpg = jp.docking.VinaPoseGenerator(prepare_protein=True)
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, ligand),
                                 centroid=(80, 38, 48),
                                 box_dims=(21, 21, 29),
                                 exhaustiveness=18,
                                 num_modes=1,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    # Check only one output since num_modes==1
    docked_outputs = list(docked_outputs)
    print(docked_outputs)
    print(len(docked_outputs))
    print(len(docked_outputs[0]))
    assert len(docked_outputs) == 1
    assert len(docked_outputs[0]) == 2





  # @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_docker_pose_generator_scores(self):
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=4,
                                 num_modes=4,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    # Check only one output since num_modes==1
    docked_outputs = list(docked_outputs)
    print(docked_outputs)
    print(len(docked_outputs))
    print(len(docked_outputs[0]))
    assert len(docked_outputs) == 4
    assert len(docked_outputs[0]) == 2

  @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_docker_pose_generator_scores_on_pocket(self):
    """Test that Docker can get scores from pose_generator."""
    # We provide no scoring model so the docker won't score
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=40,
                                 num_modes=4,
                                 out_dir="./tmp",
                                 use_pose_generator_scores=True)

    # Check only one output since num_modes==1
    docked_outputs = list(docked_outputs)
    print(docked_outputs)
    assert len(docked_outputs) == 1
    assert len(docked_outputs[0]) == 2

  @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_docker_specified_pocket(self):
    """Test that Docker can dock into spec. pocket."""
    # Let's turn on logging since this test will run for a while
    logging.basicConfig(level=logging.INFO)
    vpg = jp.docking.VinaPoseGenerator()
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 centroid=(10, 10, 10),
                                 box_dims=(10, 10, 10),
                                 exhaustiveness=1,
                                 num_modes=1,
                                 out_dir="./tmp")

    # Check returned files exist
    assert len(list(docked_outputs)) == 1

  @unittest.skip("skipping automated test")
  @unittest.skipIf(IS_WINDOWS, "vina is not supported in windows")
  @pytest.mark.slow
  def test_pocket_docker_dock(self):
    """Test that Docker can find pockets and dock dock."""
    # Let's turn on logging since this test will run for a while
    logging.basicConfig(level=logging.INFO)
    pocket_finder = jp.docking.ConvexHullPocketFinder()
    vpg = jp.docking.VinaPoseGenerator(pocket_finder=pocket_finder)
    docker = jp.docking.Docker(vpg)
    docked_outputs = docker.dock((self.protein_file, self.ligand_file),
                                 exhaustiveness=1,
                                 num_modes=1,
                                 num_pockets=1,
                                 out_dir="./tmp")

    # Check returned files exist
    assert len(list(docked_outputs)) == 1

  @pytest.mark.slow
  def test_scoring_model_and_featurizer(self):
    """Test that scoring model and featurizer are invoked correctly."""

    class DummyFeaturizer(ComplexFeaturizer):

      def featurize(self, complexes, *args, **kwargs):
        return np.zeros((len(complexes), 5))

    class DummyModel(Model):

      def predict(self, dataset, *args, **kwargs):
        return np.zeros(len(dataset))

    class DummyPoseGenerator(PoseGenerator):

      def generate_poses(self, *args, **kwargs):
        return [None]

    featurizer = DummyFeaturizer()
    scoring_model = DummyModel()
    pose_generator = DummyPoseGenerator()
    docker = jp.docking.Docker(pose_generator, featurizer, scoring_model)
    outputs = docker.dock(None)
    assert list(outputs) == [(None, np.array([0.]))]


import numpy as np
from onnx import ValueInfoProto, TensorProto

from skl2onnx.algebra.type_helper import guess_initial_types
from skl2onnx.common.data_types import (
    guess_numpy_type,
    DoubleTensorType,DataType,
    BooleanTensorType,
    Complex128TensorType, Complex64TensorType,
    FloatTensorType,
    Int8TensorType, Int16TensorType, Int32TensorType, Int64TensorType,
    UInt8TensorType, UInt16TensorType, UInt32TensorType, UInt64TensorType,
    StringTensorType
)

def select_ONNX_type(t):
    if 'float' in t:
        return FloatTensorType([None, 1])
    elif t == 'bool':
        return BooleanTensorType([None, 1])
    elif t == 'complex128':
        return Complex128TensorType([None, 1])
    elif t == 'complex64':
        return Complex64TensorType([None, 1])
    elif t == 'int8':
        return Int8TensorType([None, 1])
    elif t == 'int16':
        return Int16TensorType([None, 1])
    elif t == 'int32':
        return Int32TensorType([None, 1])
    elif t == 'int64':
        return Int64TensorType([None, 1])
    elif t == 'uint8':
        return UInt8TensorType([None, 1])
    elif t == 'uint16':
        return UInt16TensorType([None, 1])
    elif t == 'uint32':
        return UInt32TensorType([None, 1])
    elif t == 'uint64':
        return UInt64TensorType([None, 1])
    elif t == 'str':
        return StringTensorType([None, 1])
    else:
        raise ValueError('"{}" is not a valid datatype'.format(t))
    # return None


def guess_ONNX_types(df, drop=None):
    if drop is None:
        df_copy = df.copy()
    else:
        df_copy = df.drop(drop, axis=1).copy()

    type_list = [(k, select_ONNX_type(str(v)))for k, v in zip(df_copy.columns, df_copy.dtypes)]
    return type_list
    

    










#
# def guess_dtype(proto_type):
#     """
#     Converts a proto type into a :epkg:`numpy` type.
#     :param proto_type: example ``onnx.TensorProto.FLOAT``
#     :return: :epkg:`numpy` dtype
#     """
#     if proto_type == TensorProto.FLOAT:  # pylint: disable=E1101
#         return np.float32
#     if proto_type == TensorProto.BOOL:  # pylint: disable=E1101
#         return np.bool_
#     if proto_type == TensorProto.DOUBLE:  # pylint: disable=E1101
#         return np.float64
#     if proto_type == TensorProto.STRING:  # pylint: disable=E1101
#         return np.str_
#     if proto_type == TensorProto.INT64:  # pylint: disable=E1101
#         return np.int64
#     if proto_type == TensorProto.INT32:  # pylint: disable=E1101
#         return np.int32
#     if proto_type == TensorProto.INT8:  # pylint: disable=E1101
#         return np.int8
#     if proto_type == TensorProto.INT16:  # pylint: disable=E1101
#         return np.int16
#     if proto_type == TensorProto.UINT64:  # pylint: disable=E1101
#         return np.uint64
#     if proto_type == TensorProto.UINT32:  # pylint: disable=E1101
#         return np.uint32
#     if proto_type == TensorProto.UINT8:  # pylint: disable=E1101
#         return np.uint8
#     if proto_type == TensorProto.UINT16:  # pylint: disable=E1101
#         return np.uint16
#     if proto_type == TensorProto.FLOAT16:  # pylint: disable=E1101
#         return np.float16
#     raise ValueError(
#         f"Unable to convert proto_type {proto_type} to numpy type.")

#
# def guess_ONNX_types(X, itype, dtype):
#     initial_types = guess_initial_types(X, itype)
#     if dtype is None:
#         if hasattr(X, 'dtypes'):  # DataFrame
#             dtype = np.float32
#         elif hasattr(X, 'dtype'):
#             dtype = X.dtype
#         elif hasattr(X, 'type'):
#             dtype = guess_numpy_type(X.type)
#         elif isinstance(initial_types[0], ValueInfoProto):
#             dtype = guess_dtype(initial_types[0].type.tensor_type.elem_type)
#         elif initial_types is not None:
#             dtype = guess_numpy_type(initial_types[0][1])
#         else:
#             raise RuntimeError(  # pragma: no cover
#                 f"dtype cannot be guessed: {type(X)}")
#         if dtype != np.float64:
#             dtype = np.float32
#     if dtype is None:
#         raise RuntimeError("dtype cannot be None")  # pragma: no cover
#     if isinstance(dtype, FloatTensorType):
#         dtype = np.float32  # pragma: no cover
#     elif isinstance(dtype, DoubleTensorType):
#         dtype = np.float64  # pragma: no cover
#     new_dtype = dtype
#     if isinstance(dtype, np.ndarray):
#         new_dtype = dtype.dtype  # pragma: no cover
#     elif isinstance(dtype, DataType):
#         new_dtype = np.float32  # pragma: no cover
#     if new_dtype not in (np.float32, np.float64, np.int64,
#                          np.int32, np.float16):
#         raise NotImplementedError(  # pragma: no cover
#             f"dtype should be real not {new_dtype} ({dtype})")
#     return initial_types, dtype, new_dtype



from abc import ABC
import pandas as pd
import numpy as np
from typing import Iterable, Any
import math
from jaqpotpy.descriptors.molecular import RDKitDescriptors, MordredDescriptors
import pickle
# import dill

def calculate_a(X):
    shape = X.shape
    a = (3 * (shape[1] + 1)) / shape[0]
    return a


def calculate_doa_matrix(X):
    x_T = X.transpose()
    x_out = x_T.dot(X)
    x_out_inv = pd.DataFrame(np.linalg.pinv(x_out.values), x_out.columns, x_out.index)
    return x_out_inv


def calc_doa(doa_matrix, new_data):
    doaAll = []
    for nd in new_data:
        d1 = np.dot(nd, doa_matrix)
        ndt = np.transpose(nd)
        d2 = np.dot(d1, ndt)
        doa = {'DOA': d2}
        doaAll.append(doa)
    return doaAll


class DOA(object):
    """
    Abstract class for DOA methods
    """
    def calculate_threshold(self):
        raise NotImplementedError

    def calculate_matrix(self):
        raise NotImplementedError

    def calculate(self, data: Iterable[Any]) -> Iterable[Any]:
        raise NotImplementedError

    def fit(self, X: np.array):
        raise NotImplementedError

    def predict(self, data: Iterable[Any]) -> Iterable[Any]:
        raise NotImplementedError


class Leverage(DOA, ABC):
    """
    Implements DOA method leverage.
    Initialized upon training data and holds the doa matrix and the threshold 'A' value.
    Calculates the DOA for a new instance of data or array of data.
    """
    _doa = []
    _in = []

    @property
    def __name__(self):
        return 'LeverageDoa'

    def __init__(self) -> None:
        # self._scaler: BaseEstimator = scaler
        self._data: np.array = None
        self._doa_matrix = None
        self._a = None

    def __getitem__(self):
        return self

    @property
    def doa_new(self):
        return self._doa

    @doa_new.setter
    def doa_new(self, value):
        self._doa = value

    @property
    def IN(self):
        return self._in

    @IN.setter
    def IN(self, value):
        self._in = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def doa_matrix(self):
        return self._doa_matrix

    @doa_matrix.setter
    def doa_matrix(self, value):
        self._doa_matrix = value

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, value):
        self._a = value

    def calculate_threshold(self):
        shape = self._data.shape
        a = (3 * (shape[1] + 1)) / shape[0]
        self._a = a

    def calculate_matrix(self):
        x_T = self._data.transpose()
        # for s in self._data:
        #     print(s)
        #     if isinstance(s, str):
        #         print("STRING")
        x_out = x_T.dot(self._data)
        self._doa_matrix = np.linalg.pinv(x_out)
        # self.doa_matrix = x_out #pd.DataFrame(np.linalg.pinv(x_out.values), x_out.columns, x_out.index)

    def fit(self, X: np.array):
        # self._scaler.fit(X)
        # self._data = self._scaler.transform(X)
        self._data = X
        self.calculate_matrix()
        self.calculate_threshold()

    def predict(self, new_data: np.array) -> Iterable[Any]:
        doaAll = []
        self._doa = []
        self._in = []
        # new_data = self._scaler.transform(new_data)
        for nd in new_data:
            d1 = np.dot(nd, self.doa_matrix)
            ndt = np.transpose(nd)
            d2 = np.dot(d1, ndt)
            if d2 < self._a:
                in_ad = True
            else:
                in_ad = False
            self._doa.append(d2)
            self._in.append(in_ad)
            doa = {'DOA': d2, 'A': self._a, 'IN': in_ad}
            doaAll.append(doa)
        return doaAll


class MeanVar(DOA, ABC):
    """
    Implements Mean and Variance domain of applicability .
    Initialized upon training data and holds the doa mean and the variance of the data.
    Calculates the mean and variance for a new instance of data or array of data and decides if in AD.
    """
    _doa = []
    _in = []

    @property
    def __name__(self):
        return 'MeanVar'

    def __init__(self) -> None:
        # self._scaler: BaseEstimator = scaler
        self._data: np.array = None
        self._doa_matrix = None
        self._a = None

    @property
    def doa_new(self):
        return self._doa

    @doa_new.setter
    def doa_new(self, value):
        self._doa = value

    @property
    def IN(self):
        return self._in

    @IN.setter
    def IN(self, value):
        self._in = value

    @property
    def doa_matrix(self):
        return self._doa_matrix

    @doa_matrix.setter
    def doa_matrix(self, value):
        self._doa_matrix = value

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, value):
        self._a = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    def fit(self, X: np.array):
        self._data = X
        # self._scaler.fit(X)
        # self._data = self._scaler.transform(X)
        columns = list(zip(*self._data))
        shape = X.shape
        list_m_var = []
        for i in range(shape[1]):
            list_m_var.append([np.mean(columns[i]), np.std(columns[i]), np.var(columns[i])])
        self._data = np.array(list_m_var)
        self._doa_matrix = np.array(list_m_var)
        self._a = np.array(list_m_var)

    def calculate(self, new_data: np.array) -> Iterable[Any]:
        doaAll = []
        self._doa = []
        self._in = []
        # new_data = self._scaler.transform(new_data)
        in_doa = True
        for nd in new_data:
            for index, row in enumerate(nd):
                bounds = self._data[index]
                bounds_data = [bounds[0]-4*bounds[1], bounds[0]+4*bounds[1]]
                if row >= bounds_data[0] and row <= bounds_data[1]:
                    continue
                else:
                    in_doa = False
            # if len(new_data[0]) > 100 and many > 5:
            #     in_doa = False
            doa = {'IN': in_doa}
            doaAll.append(doa)
        return doaAll

    def predict(self, new_data: np.array) -> Iterable[Any]:
        doaAll = []
        self._doa = []
        self._in = []
        # new_data = self._scaler.transform(new_data)
        in_doa = True
        for nd in new_data:
            for index, row in enumerate(nd):
                bounds = self._data[index]
                bounds_data = [bounds[0]-4*bounds[1], bounds[0]+4*bounds[1]]
                if row >= bounds_data[0] and row <= bounds_data[1]:
                    continue
                else:
                    in_doa = False
            # if len(new_data[0]) > 100 and many > 5:
            #     in_doa = False
            doa = {'IN': in_doa}
            doaAll.append(doa)
            self._doa.append(new_data)
            self._in.append(in_doa)
        return doaAll


class SmilesLeverage(DOA, ABC):
    """
    Implements DOA method leverage given an array of smiles.
    Descriptors and data matrix is calculated with rdkit descriptors
    Initialized upon training data and holds the doa matrix and the threshold 'A' value.
    Calculates the DOA for a new instance of data or array of data.
    """
    _doa = []
    _in = []

    @property
    def __name__(self):
        return 'SmilesLeverage'

    def __init__(self) -> None:
        # self._scaler: BaseEstimator = scaler
        self._smiles = None
        self._data: np.array = None
        self._doa_matrix = None
        self._a = None
        # self.featurizer = MordredDescriptors(ignore_3D=True)
        self.featurizer = RDKitDescriptors(use_fragment=False, ipc_avg=False)

    def __getitem__(self):
        return self

    @property
    def doa_new(self):
        return self._doa

    @doa_new.setter
    def doa_new(self, value):
        self._doa = value

    @property
    def IN(self):
        return self._in

    @IN.setter
    def IN(self, value):
        self._in = value

    @property
    def smiles(self):
        return self._smiles

    @smiles.setter
    def smiles(self, value):
        self._smiles = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def doa_matrix(self):
        return self._doa_matrix

    @doa_matrix.setter
    def doa_matrix(self, value):
        self._doa_matrix = value

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, value):
        self._a = value

    def calculate_threshold(self):
        shape = self._data.shape
        a = (3 * (shape[1] + 1)) / shape[0]
        self._a = a

    def calculate_matrix(self):
        x_T = self._data.transpose()
        x_out = x_T.dot(self._data)
        self._doa_matrix = np.linalg.pinv(x_out)
        # self.doa_matrix = x_out #pd.DataFrame(np.linalg.pinv(x_out.values), x_out.columns, x_out.index)

    def fit(self, smiles: Iterable[str]):
        # self._scaler.fit(X)
        # self._data = self._scaler.transform(X)
        self._smiles = smiles
        from jaqpotpy.descriptors.molecular import RDKitDescriptors
        featurizer = RDKitDescriptors(use_fragment=False, ipc_avg=False)
        self._data = featurizer.featurize(smiles)
        self.calculate_matrix()
        self.calculate_threshold()

    def predict(self, smiles: Iterable[str]) -> Iterable[Any]:
        doaAll = []
        self._doa = []
        self._in = []
        # new_data = self._scaler.transform(new_data)
        from jaqpotpy.descriptors.molecular import RDKitDescriptors
        featurizer = RDKitDescriptors(use_fragment=False, ipc_avg=False)
        new_data = featurizer.featurize(smiles)
        # new_data = self.featurizer.featurize(smiles)
        for nd in new_data:
            d1 = np.dot(nd, self.doa_matrix)
            ndt = np.transpose(nd)
            d2 = np.dot(d1, ndt)
            if d2 < self._a:
                in_ad = True
            else:
                in_ad = False
            self._doa.append(d2)
            self._in.append(in_ad)
            doa = {'DOA': d2, 'A': self._a, 'IN': in_ad}
            doaAll.append(doa)
        return doaAll



from __future__ import annotations
from typing import Generic, Iterable, Mapping, Optional, Sequence, Type, TypeVar
import numpy as np
from jaqpotpy.utils import seeding
from jaqpotpy.utils.types import RDKitMol

T_cov = TypeVar("T_cov", covariant=True)


class Action(Generic[T_cov]):
    """Define the action for molecule optimization, through Actions the Jaqpotpy env actions are formulated
    and applied to any Jaqpotpy Env.
    """

    def __init__(self, action, action_meta):
        self._action = action
        self._action_meta = action_meta

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, value):
        self._action = value

    @property
    def action_meta(self):
        return self._action_meta

    @action_meta.setter
    def action_meta(self, value):
        self._action_meta = value


class Actions(Generic[T_cov]):
    """Defines the actions available for molecule optimization, so you can write actions
    code that applies to any Jaqpotpy Env.
    """
    @property
    def actions(self):
        return self.actions

    @actions.setter
    def actions(self, value: Iterable[Action]):
        self.actions = value


class Space(Generic[T_cov]):
    """Defines the observation and action spaces, so you can write generic
    code that applies to any Env. For example, you can choose a random
    action.
    WARNING - Custom observation & action spaces can inherit from the `Space`
    class. However, most use-cases should be covered by the existing space
    classes (e.g. `Box`, `Discrete`, etc...), and container classes (`Tuple` &
    `Dict`). Note that parametrized probability distributions (through the
    `sample()` method), and batching functions (in `gym.vector.VectorEnv`), are
    only well-defined for instances of spaces provided in gym by default.
    Moreover, some implementations of Reinforcement Learning algorithms might
    not handle custom spaces properly. Use custom spaces with care.
    """

    def __init__(
        self,
        shape: Optional[Sequence[int]] = None,
        dtype: Optional[Type | str] = None,
        seed: Optional[int] = None,
    ):
        self._shape = None if shape is None else tuple(shape)
        self.dtype = None if dtype is None else np.dtype(dtype)
        self._np_random = None
        if seed is not None:
            self.seed(seed)

    @property
    def np_random(self) -> seeding.RandomNumberGenerator:
        """Lazily seed the rng since this is expensive and only needed if
        sampling from this space.
        """
        if self._np_random is None:
            self.seed()

        return self._np_random  # type: ignore  ## self.seed() call guarantees right type.

    @property
    def shape(self) -> Optional[tuple[int, ...]]:
        """Return the shape of the space as an immutable property"""
        return self._shape

    def sample(self) -> T_cov:
        """Randomly sample an element of this space. Can be
        uniform or non-uniform sampling based on boundedness of space."""
        raise NotImplementedError

    def seed(self, seed: Optional[int] = None) -> list:
        """Seed the PRNG of this space."""
        self._np_random, seed = seeding.np_random(seed)
        return [seed]

    def contains(self, x) -> bool:
        """
        Return boolean specifying if x is a valid
        member of this space
        """
        raise NotImplementedError

    def __contains__(self, x) -> bool:
        return self.contains(x)

    def __setstate__(self, state: Iterable | Mapping):
        # Don't mutate the original state
        state = dict(state)

        # Allow for loading of legacy states.
        # See:
        #   https://github.com/openai/gym/pull/2397 -- shape
        #   https://github.com/openai/gym/pull/1913 -- np_random
        #
        if "shape" in state:
            state["_shape"] = state["shape"]
            del state["shape"]
        if "np_random" in state:
            state["_np_random"] = state["np_random"]
            del state["np_random"]

        # Update our state
        self.__dict__.update(state)

    def to_jsonable(self, sample_n: Sequence[T_cov]) -> list:
        """Convert a batch of samples from this space to a JSONable data type."""
        # By default, assume identity is JSONable
        return list(sample_n)

    def from_jsonable(self, sample_n: list) -> list[T_cov]:
        """Convert a JSONable data type to a batch of samples from this space."""
        # By default, assume identity is JSONable
        return sample_n


class Discrete(Space):
    r"""A discrete space in :math:`\{ 0, 1, \\dots, n-1 \}`.
    Example::
        >>> Discrete(2)
    """
    def __init__(self, n):
        assert n >= 0
        self.n = n
        super(Discrete, self).__init__((), np.int64)

    def sample(self):
        return self.np_random.randint(self.n)

    def contains(self, x):
        if isinstance(x, int):
            as_int = x
        elif isinstance(x, (np.generic, np.ndarray)) and (x.dtype.char in np.typecodes['AllInteger'] and x.shape == ()):
            as_int = int(x)
        else:
            return False
        return as_int >= 0 and as_int < self.n

    def __repr__(self):
        return "Discrete(%d)" % self.n

    def __eq__(self, other):
        return isinstance(other, Discrete) and self.n == other.n



from .torch_geometric import GCN_V1, GhebConvV1\
    , SAGEConv_V1, GraphConv_V1, GatedGraphConv_V1\
    , ResGatedGraphConv_V1, GATConv_V1, AttentiveFP_V1, AttentiveFPModel_V1, AttentiveFP
from .torch_vision import CNN
from .torch import Feedforward_V1, LSTM_V1, RNN_V1



from jaqpotpy.cfg import config
import torch
import torch.nn.functional as F
from torch.nn import Linear

class Feedforward_V1(torch.nn.Module):
    def __init__(self, input_size
                 , num_layers, hidden_layers, out_size
                 , activation: torch.nn.Module = torch.nn.ReLU(), dropout=0.0, norm: torch.nn.Module = None, act_first=True):
        super(Feedforward_V1, self).__init__()
        torch.manual_seed(config.global_seed)

        self.act = activation
        self.dropout = dropout
        self.act_first = act_first
        self.norm = norm

        self.layers = torch.nn.ModuleList([torch.nn.Linear(input_size, hidden_layers)]) #.jittable()
        for i in range(num_layers - 1):
            self.layers.append(Linear(hidden_layers, hidden_layers)) # .jittable()
        self.out = torch.nn.Linear(hidden_layers, out_size)

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
            if self.norm is not None and self.act_first is True:
                x = self.act(x)
                x = self.norm(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            elif self.norm is not None and self.act_first is False:
                x = self.norm(x)
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            else:
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.out(x)
        return x


class LSTM_V1(torch.nn.Module):
    def __init__(self, input_size, hidden_layer_size, out_size,
                 activation: torch.nn.Module = torch.nn.ReLU(), dropout=0.0, norm: torch.nn.Module = None, act_first=True):
        super(LSTM_V1, self).__init__()
        torch.manual_seed(config.global_seed)

        self.act = activation
        self.dropout = dropout
        self.act_first = act_first
        self.norm = norm
        self.hidden_layer_size = hidden_layer_size

        self.lstm = torch.nn.LSTM(input_size, hidden_layer_size)

        self.out = torch.nn.Linear(hidden_layer_size, out_size)

        self.hidden_cell = (torch.zeros(1, 1, self.hidden_layer_size),
                            torch.zeros(1, 1, self.hidden_layer_size))

    def forward(self, x):
        if self.norm is not None and self.act_first is True:
            x = self.act(x)
            x = self.norm(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        elif self.norm is not None and self.act_first is False:
            x = self.norm(x)
            x = self.act(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        else:
            x = self.act(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        lstm_out, self.hidden_cell = self.lstm(x.view(len(x) ,1, -1), self.hidden_cell)
        x = self.out(lstm_out.view(len(x), -1))
        return x


class RNN_V1(torch.nn.Module):
    def __init__(self, input_size
                 , num_layers, hidden_layers, out_size
                 , activation: torch.nn.Module = torch.nn.ReLU(), dropout=0.0, norm: torch.nn.Module = None, act_first=True):
        super(RNN_V1, self).__init__()
        torch.manual_seed(config.global_seed)

        self.act = activation
        self.dropout = dropout
        self.act_first = act_first
        self.norm = norm

        # Number of hidden dimensions
        self.hidden_dim = num_layers

        # Number of hidden layers
        self.layer_dim = hidden_layers

        # RNN
        self.rnn = torch.nn.RNN(input_size, self.hidden_dim, self.layer_dim, batch_first=True, nonlinearity='relu')

        # Readout layer
        self.out = torch.nn.Linear(self.hidden_dim, out_size)

    def forward(self, x):
        if self.norm is not None and self.act_first is True:
            x = self.act(x)
            x = self.norm(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        elif self.norm is not None and self.act_first is False:
            x = self.norm(x)
            x = self.act(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        else:
            x = self.act(x)
            x = F.dropout(x, p=self.dropout, training=self.training)

        # Initialize hidden state with zeros
        h0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim)
        # One time step
        x, hn = self.rnn(x, h0)
        x = self.out(x[:, -1, :])
        return x




  
from pydantic import BaseModel
from typing import List, Optional

class MetaInfo(BaseModel):
    identifiers: Optional[List[str]]
    comments: Optional[List[str]]
    descriptions: Optional[List[str]]
    titles: Optional[List[str]]
    subjects: Optional[List[str]]
    publishers: Optional[List[str]]
    creators: Optional[List[str]]
    contributors: Optional[List[str]]
    audiences: Optional[List[str]]
    rights: Optional[List[str]]
    sameAs: Optional[List[str]]
    seeAlso: Optional[List[str]]
    hasSources: Optional[List[str]]
    doi: Optional[List[str]]
    date: Optional[date]
    picture: Optional[str]
    markdown: Optional[str]
    tags: Optional[List[str]]
    read: Optional[List[str]]
    write: Optional[List[str]]
    execute: Optional[List[str]]

class Doa(BaseModel):
    meta: Optional[MetaInfo]
    modelId: Optional[str]
    doaMatrix: Optional[List[List[float]]]
    aValue: Optional[float]




from jaqpotpy.datasets.dataset_base import MolecularDataset
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
try:
    from jaqpotpy.descriptors.molecular import MolGraphConvFeaturizer, TorchMolGraphConvFeaturizer
except ImportError:
    pass
from typing import Iterable, Any
import pandas as pd
import os
import inspect
import numpy as np
import pickle
from torch.utils.data import Dataset


class MolecularTabularDataset(MolecularDataset):
    """
    Reads CSV with smiles, experimental features and endpoints
    """
    def __init__(self, path, x_cols=None, y_cols=Iterable[Any], smiles_col=None, smiles=Iterable[str]
                 , y: Iterable[Any] = Iterable[Any], X: Iterable[Any] = None,
                 featurizer: MolecularFeaturizer = None, task: str = "regression") -> None:
        # super(SmilesTabularDataset, path, x_cols, y_cols).__init__(path, x_cols, y_cols)
        super(MolecularTabularDataset, self).__init__(path=path, x_cols=x_cols, y_cols=y_cols)
        self._y = y
        self._x = X
        self._df = None
        self._external = None
        self._x_cols_all = None
        self.smiles_col = smiles_col
        self.smiles = smiles
        self._smiles_strings = None
        self.ys = y
        self._task = task
        self.featurizer: MolecularFeaturizer = featurizer
        self.indices: [] = None
        # self.create()

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdb", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpotpy_dataset" + ".jdb", 'wb') as f:
                pickle.dump(self, f)

    def create(self):
        name, extension = os.path.splitext(self.path)
        if extension == '.csv':
            data = pd.read_csv(self.path)
            smiles = data[self.smiles_col].to_list()
            self._smiles_strings = smiles
            self.smiles = smiles
            descriptors = self.featurizer.featurize_dataframe(smiles)
            # print(len(list(descriptors)))
            if self.x_cols:
                self._external = self.x_cols
                xs = data[self.x_cols]
                y = data[self.y_cols]
                self.df = pd.concat([descriptors, xs, y], axis=1)
                self._x_cols_all = list(descriptors) + list(xs)
                self.y_cols = list(y)
                self.y = list(y)
            else:
                y = data[self.y_cols]
                self.df = pd.concat([descriptors, y], axis=1)
                self._x_cols_all = list(descriptors)
                self.y_cols = list(y)
                self.y = list(y)
            if self.X is None:
                if self.x_cols:
                    xs = data[self.x_cols]
                    if not xs.empty:
                        self._x = list(descriptors) + list(xs)
                    else:
                        self._x = list(descriptors)
                else:
                    self._x = list(descriptors)
        return self

    def __get_X__(self):
        return self.df[self.X].to_numpy()

    def __get_Y__(self):
        return self.df[self.y].to_numpy()

    def __get__(self):
        return self.df

    def __getitem__(self, idx):
        # print(self.df[self.X].iloc[idx].values)
        # print(type(self.df[self.X].iloc[idx].values))
        X = self.df[self.X].iloc[idx].values
        y = self.df[self.y].iloc[idx].to_numpy()
        return X, y

    def __len__(self):
        if self.df is None:
            self.create()
        return len(self.df)

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
        #   # for list
        return self.__class__.__name__


class TorchGraphDataset(MolecularDataset, Dataset):
    """
    Init with smiles and y array
    """
    def __init__(self, smiles=Iterable[str]
                 , y: Iterable[Any] = Iterable[Any], featurizer: MolecularFeaturizer = MolGraphConvFeaturizer(), task=str, streaming: bool = False) -> None:
        super(TorchGraphDataset, self).__init__()
        self._y = y
        self._dataset_name = None
        self._df = []
        self._x_cols_all = None
        self.smiles = smiles
        self._smiles_strings = None
        self.ys = y
        self._task = task
        self.featurizer: MolecularFeaturizer = featurizer
        self.indices: [] = None
        self.streaming = streaming
        # self.create()

    def create(self):
        if self.streaming is False:
            import torch
            from torch_geometric.data import Data
            descriptors = self.featurizer.featurize(datapoints = self.smiles)
            for i, g in enumerate(descriptors):
                if self._task == 'regression':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr= torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.Tensor([self.y[i]]))
                    self._df.append(dato)
                elif self._task == 'classification':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr=torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.LongTensor([self.y[i]]))
                    self._df.append(dato)
                else:
                    raise Exception("Please set task (classification / regression).")
            self._smiles_strings = self.smiles
            self.X = ['TorchMolGraph']
            self.y = ['Y']
            return self
        else:
            ys = pd.DataFrame(self.ys, columns=['Y'])
            self.df = pd.concat([self.smiles, ys], axis=1)
            return self

    def __get__(self):
        return self.df

    def __getitem__(self, idx):
        if self.streaming is False:
            return self.df[idx]
        else:
            tempdf = []
            import torch
            from torch_geometric.data import Data
            smile = self.smiles[idx]
            descriptors = self.featurizer.featurize(smile)
            for i, g in enumerate(descriptors):
                if self._task == 'regression':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr= torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.Tensor([self.y[i]]))
                    tempdf.append(dato)
                elif self._task == 'classification':
                    dato = Data(x=torch.FloatTensor(g.node_features)
                                , edge_index=torch.LongTensor(g.edge_index)
                                , edge_attr=torch.LongTensor(g.edge_features) if g.edge_features is not None else g.edge_features
                                , num_nodes=g.num_nodes, y=torch.LongTensor([self.y[i]]))
                    tempdf.append(dato)
                else:
                    raise Exception("Please set task (classification / regression).")
            self._smiles_strings = self.smiles
            self.X = ['TorchMolGraph']
            self.y = ['Y']
            return tempdf[idx]

    def __len__(self):
        return len(self.df)

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
          # for list
        return self.__class__.__name__


class SmilesDataset(MolecularDataset):
    """
    Init with smiles and y array
    """
    def __init__(self, x_cols=None, smiles=Iterable[str]
                 , y: Iterable[Any] = Iterable[Any], X: Iterable[Any] = None,
                 featurizer: MolecularFeaturizer = None, task: str = 'regression', streaming: bool = False) -> None:
        super(SmilesDataset, self).__init__(x_cols=x_cols)
        self._y = y
        self._x = X
        self._dataset_name = None
        self._df = None
        self._x_cols_all = None
        self.smiles = smiles
        self._smiles_strings = None
        self._external = None
        self.ys = y
        self.featurizer: MolecularFeaturizer = featurizer
        self.indices: [] = None
        self._task = task
        self.streaming = streaming

        # self.create()

    def create(self):
        if self.streaming is False:
            descriptors = self.featurizer.featurize_dataframe(self.smiles)
            if self._task != "generation":
                ys = pd.DataFrame(self.ys, columns=['Y'])
                self.df = pd.concat([descriptors, ys], axis=1)
                self._smiles_strings = self.smiles
                self.X = list(descriptors)
                self.y = ['Y']
            else:
                self.df = pd.concat([descriptors], axis=1)
                self._smiles_strings = self.smiles
                self.X = list(descriptors)
            return self
        else:
            ys = pd.DataFrame(self.ys, columns=['Y'])
            self.df = pd.concat([self.smiles, ys], axis=1)
            return self

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdb", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpotpy_dataset" + ".jdb", 'wb') as f:
                pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def __get_X__(self):
        return self.df[self.X].to_numpy()

    def __get_Y__(self):
        return self.df[self.y].to_numpy()

    def __get__(self):
        return self.df

    def __getitem__(self, idx):
        if self.streaming is False:
            if self._task != "generation":
                if self.X == ['Sequence']:
                    X = self.df[self.X].iloc[idx].values[0]
                elif self.X == ['OneHotSequence']:
                    X = self.df[self.X].iloc[idx].values[0]
                elif self.X == ['SmilesImage']:
                    X = self.df[self.X].iloc[idx].values[0].transpose(2, 0, 1)
                else:
                    X = self.df[self.X].iloc[idx].values
                y = self.df[self.y].iloc[idx].to_numpy()
                return X, y
            else:
                try:
                    node_features = self.df[self.X].iloc[idx].values[0].node_features
                    adjacency_matrix = self.df[self.X].iloc[idx].values[0].adjacency_matrix
                    import torch
                    X = torch.Tensor(node_features), torch.Tensor(adjacency_matrix)
                except Exception as e:
                    print(idx)
                    print(self.df[self.X].iloc[idx])
                    print(str(e))
                return X, idx, (torch.Tensor(node_features), torch.Tensor(adjacency_matrix))
        elif self.streaming is True:
            smiles = self.smiles[idx]
            descriptors = self.featurizer.featurize_dataframe(smiles)
            if self._task != "generation":
                y = self.ys[idx]
                ys = pd.DataFrame([y], columns=['Y'])
                temp_df = pd.concat([descriptors, ys], axis=1)
                self._smiles_strings = self.smiles
                self.X = list(descriptors)
                self.y = ['Y']
            else:
                temp_df = pd.concat([descriptors], axis=1)
                self.X = list(descriptors)
            if self._task != "generation":
                if self.X == ['Sequence']:
                    X = temp_df[self.X].iloc[0].values[0]
                elif self.X == ['OneHotSequence']:
                    X = temp_df[self.X].iloc[0].values[0]
                elif self.X == ['SmilesImage']:
                    X = temp_df[self.X].iloc[0].values[0].transpose(2, 0, 1)
                else:
                    X = temp_df[self.X].iloc[0].values
                y = temp_df[self.y].iloc[0].to_numpy()
                return X, y
            else:
                try:
                    node_features = temp_df[self.X].iloc[0].values[0].node_features
                    adjacency_matrix = temp_df[self.X].iloc[0].values[0].adjacency_matrix
                    import torch
                    X = torch.Tensor(node_features), torch.Tensor(adjacency_matrix)
                except Exception as e:
                    print(idx)
                    print(temp_df[self.X].iloc[0])
                    print(str(e))
                return X, idx, (torch.Tensor(node_features), torch.Tensor(adjacency_matrix))


    def __len__(self):
        if self.df is None:
            self.create()
        return len(self.df)

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
        #   # for list
        return self.__class__.__name__



from typing import List
from jaqpotpy.parsers.base_classes import Parser
from jaqpotpy.entities.material_models import (
    Xyz, Atoms
)
import pandas as pd


def _clean_key(word) -> str:
    return ''.join(i for i in word if i.isalnum())

def _str_to_num(s, start, stop, change):
    """
    string: the whole string
    start: the starting character
    stop: the stiping character
    change: the changing function --> 0: float() and 1: int()
    """

    x = s[start:stop]

    if x.strip() != '':
        if change == 0:
            x = float(x)

        else:
            x = int(x)
    return x


class XyzParser(Parser):
    """
    XyzParser.
    This class parses a xyz or a extxyz file (or all xyz or extxyz files in a folder) into a Xyz structure.
    For more information on the structure see jaqpotpy.models.material_models.Xyz

    Attributes
    ----------
    files_: str | List[str]
        File names of the pdb files that were parsed.

    Examples
    --------
    >>> import jaqpotpy as jt
    >>> xyz_file = './AgNP.xyz'
    >>> parser = jt.parsers.xyz_parser.xyz_parser.XyzParser()
    >>> xyz = parser.parse(xyz_file)
    >>> type(xyz)
    generator
    >>> parsed = next(xyz)
    >>> type(parsed)
    jaqpotpy.models.material_models.Xyz
    """

    @property
    def __name__(self):
        return 'XyzParser'

    def __getitem__(self):
        return self

    def _parse(self, path) -> Xyz:

        """
        Parse xyz or extxyz files.

        Parameters
        ----------
        path: str
          Either the path of a certain file or a path of a folder containing
          files that will be parsed

        Returns
        -------
        jaqpotpy.models.material_models.Xyz
          A Xyz object
        """

        self.files_.append(path)

        # Initialize variables
        xyz_dict: Xyz = Xyz(num_atoms=0, comment='', atoms=Atoms(elements=[], coordinates=[], extraInfo=[]))
        cnt = 0

        # Open the file and read it in as a list of rows
        with open(path) as f:
            xyz = f.read().splitlines()

        # Iterate through the file
        for row in xyz:
            if row.strip() != '':
                if cnt == 0:
                    xyz_dict.num_atoms = int(row)
                    cnt += 1
                elif cnt == 1:
                    xyz_dict.comment = row
                    cnt += 1
                else:
                    curr_list = row.split()
                    xyz_dict.atoms.elements.append(_clean_key(curr_list[0].strip()))
                    xyz_dict.atoms.coordinates.append([float(curr_list[1]), float(curr_list[2]), float(curr_list[3])])
                    try:
                        xyz_dict.atoms.extraInfo.append(curr_list[4:])
                    except:
                        pass

        return xyz_dict

    def _parse_dataframe(self, file: Xyz, filename: str) -> pd.DataFrame:
        """
        Parse xyz files in a pandas dataframe.

        Parameters
        ----------
        file: jaqpotpy.models.material_models.Xyz
            The Xyz structrure of a parsed file

        filename: str
            The name of the parsed file

        Returns
        -------
        pd.DataFrame()
        """

        df = pd.DataFrame()
        for i in range(len(file.atoms.elements)):
            d = {}
            d['file'] = filename
            d['element'] = file.atoms.elements[i]
            d['x'] = file.atoms.coordinates[i][0]
            d['y'] = file.atoms.coordinates[i][1]
            d['z'] = file.atoms.coordinates[i][2]

            df = pd.concat([df, pd.DataFrame(d, index=[0])]).reset_index(drop=True)

        return df



from jaqpotpy.descriptors.molecular.mordred import MordredDescriptors
from jaqpotpy.descriptors.molecular.smiles_to_seq import SmilesToSeq, create_char_to_idx
from jaqpotpy.descriptors.molecular.one_hot_sequence import OneHotSequence
from jaqpotpy.descriptors.molecular.rdkit import RDKitDescriptors
from jaqpotpy.descriptors.molecular.pub_chem_fingerprint import PubChemFingerprint
from jaqpotpy.descriptors.molecular.topological_fingerprints import TopologicalFingerprint
# try:
from jaqpotpy.descriptors.molecular.molecule_graph_conv import MolGraphConvFeaturizer, PagtnMolGraphFeaturizer,\
        TorchMolGraphConvFeaturizer, AttentiveFPFeaturizer
from jaqpotpy.descriptors.graph.graph_data import GraphData, BatchGraphData
from jaqpotpy.descriptors.graph.graph_features import GraphConvConstants
from jaqpotpy.descriptors.graph.mol_graphs import *
# except ModuleNotFoundError:
#     pass
from jaqpotpy.descriptors.molecular.smles_to_image import SmilesToImage
from jaqpotpy.descriptors.molecular.molgan import MolGanFeaturizer, GraphMatrix
from jaqpotpy.descriptors.molecular.maccs_keys_fingerprints import MACCSKeysFingerprint
from jaqpotpy.descriptors.molecular.coulomb_matrices import CoulombMatrix, CoulombMatrixEig




from jaqpotpy.datasets import *
from jaqpotpy.datasets.molecular_datasets import *
from jaqpotpy.datasets.image_datasets import TorchImageDataset
from jaqpotpy.datasets.molecular_datasets import MolecularDataset, MolecularTabularDataset, TorchGraphDataset
from jaqpotpy.datasets.material_datasets import CompositionDataset, StructureDataset


from jaqpotpy.models.base_classes import Model
from jaqpotpy.doa.doa import DOA
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
from typing import Any, Union
import pandas as pd
import pickle
from jaqpotpy.datasets import MolecularTabularDataset, Dataset, TorchGraphDataset, SmilesDataset, MolecularDataset, CompositionDataset, StructureDataset
from jaqpotpy.models import Evaluator, Preprocesses
import torch
from torch.utils.data import DataLoader
from jaqpotpy.models import MolecularModel, MaterialModel
import numpy as np
from jaqpotpy.cfg import config
import os
import jaqpotpy


class MolecularTorch(Model):

    def __init__(self, dataset: MolecularDataset, model_nn: torch.nn.Module
                 , doa: DOA = None
                 , eval: Evaluator = None, preprocess: Preprocesses = None
                 , dataLoaderParams: Any = None, epochs: int = None
                 , criterion: torch.nn.Module = None, optimizer: Any = None
                 , train_batch: int = 50, test_batch: int = 50, log_steps: int = 1, model_dir: str = "./", device: str = 'cpu', test_metric=(None, 'minimize')):
        # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
        self.dataset: MolecularDataset = dataset
        self.model_nn = model_nn
        self.doa = doa
        self.doa_m = None
        self.external = None
        self.evaluator: Evaluator = eval
        self.preprocess: Preprocesses = preprocess
        self.train_batch = train_batch
        self.test_batch = test_batch
        self.trainDataLoaderParams = {'batch_size': self.train_batch, 'shuffle': False, 'num_workers': 0}
        self.testDataLoaderParams = {'batch_size': self.test_batch, 'shuffle': False, 'num_workers': 0}
        self.epochs = epochs
        self.criterion = criterion
        self.trained_model = None
        self.model_fitted: MolecularModel = None
        self.optimizer_local = optimizer
        self.train_loader = None
        self.test_loader = None
        self.log_steps = log_steps
        self.best_model = model_nn
        self.path = None
        self.model_dir = model_dir
        self.device = torch.device(device)
        self.__default__ = False
        if test_metric[0] is None and dataset.task == 'classification':
            self.__default__ = True
            test_metric = ('Accuracy', 'maximize')
        if test_metric[0] is None and dataset.task == 'regression':
            self.__default__ = True
            test_metric = ('Loss', 'minimize')
        self.test_metric = test_metric
        # torch.multiprocessing.freeze_support()

    def __call__(self, smiles):
        self

    def fit(self):
        steps = self.epochs * 0.1
        if self.doa:
            if self.doa:
                if self.doa.__name__ == 'SmilesLeverage':
                    self.doa_m = self.doa.fit(self.dataset.smiles)
                else:
                    self.doa_m = self.doa.fit(X=self.dataset.__get_X__())
        if self.dataset.df is not None:
            pass
        else:
            self.dataset.create()
        if self.evaluator is not None:
            if self.evaluator.dataset.df is not None:
                pass
            else:
                self.evaluator.dataset.create()
        # device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.model_fitted = MolecularModel()
        self.model_nn.to(self.device)
        # self.train_loader = DataLoader(dataset=self.dataset.df, **self.trainDataLoaderParams)
        # self.test_loader = DataLoader(dataset=self.evaluator.dataset.df, **self.testDataLoaderParams)
        self.train_loader = DataLoader(dataset=self.dataset, **self.trainDataLoaderParams)
        self.test_loader = DataLoader(dataset=self.evaluator.dataset, **self.testDataLoaderParams)
        temp_loss = None
        for epoch in range(1, self.epochs):
            self.train()
            train_loss = self.test(self.train_loader)
            test_loss = self.test(self.test_loader)
            if self.dataset.task == "classification":
                los = test_loss[2]
                if self.__save_choice__(temp_loss, train_loss[2], los):
                    temp_loss = [train_loss[2], los]
                    # if temp_loss < los and steps > epoch:
                    temp_path = self.path
                    self.path = self.model_dir + "molecular_model_ep_" + str(epoch) + "_er_" + str(los) + ".pt"
                    torch.save({
                        'epoch': epoch,
                        'model_state_dict': self.model_nn.state_dict(),
                        'optimizer_state_dict': self.optimizer_local.state_dict(),
                        'loss': los,
                    }, self.path)
                    try:
                        os.remove(temp_path)
                    except TypeError as e:
                        continue
                if epoch % self.log_steps == 0:
                    print(f'Epoch: {epoch:03d}, Train Score: {train_loss[2]}, Test Score: {test_loss[2]}')
            else:
                los = test_loss[1].item()
                if self.__save_choice__(temp_loss, train_loss[2], los):
                    temp_loss = [train_loss[1], los]
                    temp_path = self.path
                    self.path = self.model_dir + "molecular_model_ep_" + str(epoch) + "_er_" + str(los) + ".pt"
                    torch.save({
                        'epoch': epoch,
                        'model_state_dict': self.model_nn.state_dict(),
                        'optimizer_state_dict': self.optimizer_local.state_dict(),
                        'loss': los,
                    }, self.path)
                    try:
                        os.remove(temp_path)
                    except TypeError as e:
                        continue
                if epoch % self.log_steps == 0:
                    print(f'Epoch: {epoch:03d}, Train Score: {train_loss[1]}, Test Score: {test_loss[1]}')
        return self

    def __save_choice__(self, temp_loss, train_loss, test_loss):
        if self.test_metric[1] == 'minimize':
            if temp_loss is None:
                return True
            elif temp_loss[1] > test_loss:
                return True
            elif temp_loss[1] == test_loss:
                if temp_loss[0] >= train_loss:
                    return True
                else:
                    return False
            else:
                return False
        elif self.test_metric[1] == 'maximize':
            if temp_loss is None:
                return True
            elif temp_loss[1] < test_loss:
                return True
            elif temp_loss[1] == test_loss:
                if temp_loss[0] <= train_loss:
                    return True
                else:
                    return False
            else:
                return False


    def train(self):
        self.model_nn.train()
        for data in self.train_loader:
            input = data[0].to(self.device)
            y = data[1].to(self.device)
            out = self.model_nn(input.float())
            if self.dataset.task == 'classification':
                truth = torch.squeeze(y.long())
                loss = self.criterion(out, truth)
            else:
                # truth = torch.squeeze(data[1].float())
                loss = self.criterion(out, y.float())
            # loss = self.criterion(out, data[1].float())
            # print(loss)
            loss.backward(retain_graph=True)
            self.optimizer_local.step()
            self.optimizer_local.zero_grad()

    def test(self, dataloader):
        if self.__default__:
            return self.__default_test__(dataloader)
        else:
            self.model_nn.eval()
            test_function = self.test_metric[0]
            truth = np.array([])
            preds = np.array([])
            if self.dataset.task == 'classification':
                for data in dataloader:
                    input = data[0].to(self.device)
                    y = data[1].to(self.device)
                    out = self.model_nn(input.float())
                    y = torch.squeeze(y.long())
                    preds = np.append(preds, out.argmax(dim=1).cpu().detach().numpy())
                    truth = np.append(truth, y)
                return out, out.argmax(dim=1).cpu().numpy(), test_function(truth, preds)
            else:
                self.model_nn.eval()
                for data in dataloader:
                    input = data[0].to(self.device)
                    y = data[0].to(self.device)
                    out = self.model_nn(input)
                    preds = np.append(preds, out.cpu().detach().numpy())
                return out, test_function(truth, preds)

    def __default_test__(self, dataloader):
        self.model_nn.eval()
        if self.dataset.task == 'classification':
            correct = 0
            for data in dataloader:
                input = data[0].to(self.device)
                y = data[1].to(self.device)
                out = self.model_nn(input.float())
                pred = out.argmax(dim=1)
                truth = torch.squeeze(y.long())
                correct += int((pred == truth).sum())
                loss = self.criterion(out, truth)
            return out, pred.cpu().numpy(), correct / len(dataloader.dataset)
        else:
            self.model_nn.eval()
            for data in dataloader:
                input = data[0].to(self.device)
                y = data[0].to(self.device)
                out = self.model_nn(input)
                loss = self.criterion(out, y.float())
            return out, loss

    def eval(self):
        checkpoint = torch.load(self.path)
        self.best_model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer_local.load_state_dict(checkpoint['optimizer_state_dict'])
        self.best_model.eval()
        if self.evaluator.functions:
            eval_keys = self.evaluator.functions.keys()
            correct = 0
            truth = np.array([])
            preds = np.array([])
            for data in self.test_loader:
                input = data[0].to(self.device)
                y = data[1].to(self.device)
                out = self.best_model(input.float())
                pred = out.argmax(dim=1)
                correct += int((pred == y.float()).sum())
                truth = np.append(truth, y.cpu().float().numpy())
                preds = np.append(preds, pred.cpu().numpy())
            # return out, pred.numpy(), correct / len(dataloader.dataset)
            for eval_key in eval_keys:
                eval_function = self.evaluator.functions.get(eval_key)
                print(eval_key + ": " + str(eval_function(truth, preds)))
        else:
            print("No eval functions passed")

    def create_molecular_model(self):
        checkpoint = torch.load(self.path)
        self.best_model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer_local.load_state_dict(checkpoint['optimizer_state_dict'])
        model = MolecularModel()
        if type(self.dataset.featurizer).__name__ == "RDKitDescriptors":
            model.descriptors = "RDKitDescriptors"
        else:
            model.descriptors = self.dataset.featurizer
        model.doa = self.doa
        model_s = torch.jit.script(self.best_model)
        model_save = torch.jit.save(model_s, "local_temp.pt")
        with open("./local_temp.pt", "rb") as f:
            model.model = f.read()
        f.close()
        os.remove("./local_temp.pt")
        model.optimizer = self.optimizer_local
        model.X = self.dataset.X
        model.Y = self.dataset.y
        model.library = ['torch']
        model.version = [torch.__version__]
        model.jaqpotpy_version = jaqpotpy.__version__
        model.jaqpotpy_docker = config.jaqpotpy_docker
        model.modeling_task = self.dataset.task
        # model.external_feats = self.dataset.external
        return model


class MaterialTorch(Model):

    def __init__(self, dataset: Union[CompositionDataset, StructureDataset], model_nn: torch.nn.Module
                 , doa: DOA = None
                 , eval: Evaluator = None, preprocess: Preprocesses = None
                 , dataLoaderParams: Any = None, epochs: int = None
                 , criterion: torch.nn.Module = None, optimizer: Any = None
                 , train_batch: int = 50, test_batch: int = 50, log_steps: int = 1, device: str = 'cpu'):
        # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
        self.dataset: MolecularDataset = dataset
        self.model_nn = model_nn
        self.doa = doa
        self.doa_m = None
        self.external = None
        self.evaluator: Evaluator = eval
        self.preprocess: Preprocesses = preprocess
        self.train_batch = train_batch
        self.test_batch = test_batch
        self.trainDataLoaderParams = {'batch_size': self.train_batch, 'shuffle': False, 'num_workers': 0}
        self.testDataLoaderParams = {'batch_size': self.test_batch, 'shuffle': False, 'num_workers': 0}
        self.epochs = epochs
        self.criterion = criterion
        self.trained_model = None
        self.model_fitted: MolecularModel = None
        self.optimizer_local = optimizer
        self.train_loader = None
        self.test_loader = None
        self.log_steps = log_steps
        self.best_model = None
        self.path = None
        self.device = torch.device(device)

        # torch.multiprocessing.freeze_support()

    def __call__(self, smiles):
        self

    def fit(self):
        steps = self.epochs * 0.1
        if self.doa:
            if self.doa:
                if self.doa.__name__ == 'SmilesLeverage':
                    self.doa_m = self.doa.fit(self.dataset.smiles_strings)
                else:
                    self.doa_m = self.doa.fit(X=self.dataset.__get_X__())
        if self.dataset.df is not None:
            pass
        else:
            self.dataset.create()
        if self.evaluator is not None:
            if self.evaluator.dataset.df is not None:
                pass
            else:
                self.evaluator.dataset.create()
        self.model_fitted = MolecularModel()
        self.model_nn.to(self.device)
        # self.train_loader = DataLoader(dataset=self.dataset.df, **self.trainDataLoaderParams)
        # self.test_loader = DataLoader(dataset=self.evaluator.dataset.df, **self.testDataLoaderParams)
        self.train_loader = DataLoader(dataset=self.dataset, **self.trainDataLoaderParams)
        self.test_loader = DataLoader(dataset=self.evaluator.dataset, **self.testDataLoaderParams)
        temp_loss = None
        for epoch in range(1, self.epochs):
            self.train()
            train_loss = self.test(self.train_loader)
            test_loss = self.test(self.test_loader)
            if self.dataset.task == "classification":
                los = test_loss[2]
                # if temp_loss is None or temp_loss < los and steps > epoch:
                if temp_loss is None or temp_loss < los:
                    self.best_model = self.model_nn
                if epoch % self.log_steps == 0:
                    print(f'Epoch: {epoch:03d}, Train Accuracy: {train_loss[2]}, Test Accuracy: {test_loss[2]}')
            else:
                los = test_loss[1].item()
                # if temp_loss is None or temp_loss < los and steps > epoch:
                if temp_loss is None or temp_loss > los:
                    self.best_model = self.model_nn
                if epoch % self.log_steps == 0:
                    print(f'Epoch: {epoch:03d}, Train Loss: {train_loss[1]}, Test Loss: {test_loss[1]}')
        return self

    def train(self):
        self.model_nn.train()
        for data in self.train_loader:
            data = data.to(self.device)
            out = self.model_nn(data[0].float())
            if self.dataset.task == 'classification':
                truth = torch.squeeze(data[1].long())
                truth = truth.to(self.device)
                loss = self.criterion(out, truth)
            else:
                loss = self.criterion(out, data[1].float())
            loss.backward(retain_graph=True)
            self.optimizer_local.step()
            self.optimizer_local.zero_grad()

    def test(self, dataloader):
        self.model_nn.eval()
        if self.dataset.task == 'classification':
            correct = 0
            for data in dataloader:
                input = data[0].to(self.device)
                out = self.model_nn(input.float())
                pred = out.argmax(dim=1)
                truth = torch.squeeze(data[1].long())
                correct += int((pred == truth).sum())
                loss = self.criterion(out, truth)
            return out, pred.numpy(), correct / len(dataloader.dataset)
        else:
            self.model_nn.eval()
            for data in dataloader:
                data = data.to(self.device)
                if self.best_model:
                    self.best_model.eval()
                out = self.model_nn(data[0].float())
                # truth = torch.squeeze(data[1].float())
                loss = self.criterion(out, data[1].float())
            return out, loss

    def eval(self):
        self.best_model.eval()
        if self.evaluator.functions:
            eval_keys = self.evaluator.functions.keys()
            correct = 0
            truth = np.array([])
            preds = np.array([])
            for data in self.test_loader:
                data = data.to(self.device)
                if self.dataset.task == 'classification':
                    out = self.best_model(data[0].float())
                    pred = out.argmax(dim=1)
                    correct += int((pred == data[1].float()).sum())
                    truth = np.append(truth, data[1].numpy())
                    preds = np.append(preds, pred.numpy())
                else:
                    out = self.best_model(data[0].float())
                    truth = np.append(truth, data[1].numpy())
                    preds = np.append(preds, out.detach().numpy())
            # return out, pred.numpy(), correct / len(dataloader.dataset)
            for eval_key in eval_keys:
                eval_function = self.evaluator.functions.get(eval_key)
                print(eval_key + ": " + str(eval_function(truth, preds)))
        else:
            print("No eval functions passed")

    def create_material_model(self):
        checkpoint = torch.load(self.path)
        self.best_model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer_local.load_state_dict(checkpoint['optimizer_state_dict'])
        model = MolecularModel()
        model.descriptors = self.dataset.featurizer
        model.doa = self.doa
        # model.model = self.best_model
        self.best_model.to("cpu")
        model_s = torch.jit.script(self.best_model)
        model_save = torch.jit.save(model_s, "local_temp.pt")
        with open("./local_temp.pt", "rb") as f:
            model.model = f.read()
        f.close()
        os.remove("./local_temp.pt")
        # model.model = torch.jit.script(self.best_model)
        model.optimizer = self.optimizer_local
        model.X = self.dataset.X
        model.Y = self.dataset.y
        model.library = ['torch_geometric', 'torch']
        model.version = [torch.__version__]
        model.jaqpotpy_version = jaqpotpy.__version__
        model.jaqpotpy_docker = config.jaqpotpy_docker
        model.modeling_task = self.dataset.task
        return model



from typing import Dict, Callable, Any, Iterable
from jaqpotpy.datasets.molecular_datasets import SmilesDataset


class Evaluator:
    functions: Dict[str, Callable] = {}
    dataset: Any

    def __init__(self):
        pass

    @classmethod
    def register_scoring_function(cls, function_name, function):
        cls.functions[function_name] = function

    @classmethod
    def register_dataset(cls, dataset: SmilesDataset):
        dataset = dataset

    def __getitem__(self):
        return self


class GenerativeEvaluator:
    functions: Dict[str, Callable] = {}
    eval_functions: Dict[str, Callable] = {}
    dataset: Iterable[str]

    def __init__(self):
        pass

    @classmethod
    def register_scoring_function(cls, function_name, function):
        cls.functions[function_name] = function

    @classmethod
    def register_evaluation_function(cls, function_name, function):
        cls.eval_functions[function_name] = function

    @classmethod
    def register_dataset(cls, dataset: Iterable[str]):
        cls.dataset = dataset
        # dataset = dataset

    def __getitem__(self):
        return self

    def get_reward(self, mols):
        rr = 1.
        for key in self.functions.keys():
            try:
                f = self.functions.get(key)
                rr *= f(mols)
            except TypeError as e:
                f = self.functions.get(key)
                rr *= f(mols, self.dataset)
        return rr.reshape(-1, 1)


class GenerativeReward:
    functions: Dict[str, Callable] = {}
    dataset: Any

    def __init__(self):
        pass

    @classmethod
    def register_scoring_function(cls, function_name, function):
        cls.functions[function_name] = function

    @classmethod
    def register_dataset(cls, dataset: SmilesDataset):
        dataset = dataset

    def __getitem__(self):
        return self



import os
import jaqpotpy
"""
Configuration variables for Jaqpotpy
"""

"""
x -> Test variable
verbode -> Verbose outputs. eg descriptor creation, model training
global_seed -> Seed for various descriptors and models
version -> Jaqpotpy version
"""


x = 0
verbose = True
global_seed = 42
# version = jaqpotpy.__version__
try:
    jaqpotpy_docker = os.environ['JAQPOTPY_DOCKER']
except KeyError as e:
    jaqpotpy_docker = None



from jaqpotpy.entities.feature import Feature
import math
import jaqpotpy
from jaqpotpy.cfg import config

from jaqpotpy.helpers.builders import FeatureBuilder,\
    FeatureDirector, DataEntryBuilder, DataEntryDirector,\
    PretrainedNeedsDirector, PretrainedNeedsBuilder, DoaDirector, DoaBuilder


def create_feature(feat_title, creator):
    fb = FeatureBuilder()
    fb.add_comments("Feature created from python client for dataset creation")
    fb.add_creator(creator)
    fb.add_title(feat_title)
    dire = FeatureDirector()
    mf = dire.construct(fb)
    return mf


def clear_entity(o):
    delete = []
    for key in o:
        if getattr(o, key) is None or "":
            delete.append(key)
    for keytod in delete:
        delattr(o, keytod)
    return o


def create_data_entry(df, feat_map, owner_uuid):
    data_entry = []
    for dataid in df.index:
        # data_entry_all = {}
        values_from_dataframe = df.loc[dataid]

        if type(values_from_dataframe).__name__ != 'DataFrame':
            values_from_dataframe = values_from_dataframe.to_frame()
        # values_from_dataframe = values_from_dataframe.to_json()
        # print(values_from_dataframe_j)
        # print(dataid)
        # print(values_from_dataframe)

        de_director = DataEntryDirector()
        de_builder = DataEntryBuilder()
        de_builder.set_name(dataid)
        de_builder.set_owneruuid(owner_uuid)
        values = {}
        for key in feat_map:
            values[feat_map[key]] = values_from_dataframe[dataid][key]
        de_builder.set_values(values)
        de = de_director.construct(de_builder)
        data_entry.append(de.__dict__)
    return data_entry


def create_pretrain_req(model, X, y, title, description, algorithm, implementedWith, runtime, additionalInfo):
    pnb = PretrainedNeedsBuilder()
    independentFeatures = []
    dependendFeatures = []
    if type(X).__name__ == 'DataFrame':
        for xe in list(X):
            independentFeatures.append(xe)
    elif type(X).__name__ == 'Series':
        independentFeatures.append(X.name)
    if type(y).__name__ == 'DataFrame':
        for fe in list(y):
            dependendFeatures.append(fe)
    elif type(y).__name__ == 'Series':
        dependendFeatures.append(y.name)
    pnb.setRawModel(model)
    pnb.setAdditionalInfo(additionalInfo)
    pnb.setAlgorithm(algorithm)
    pnb.setDependendFeatures(dependendFeatures)
    pnb.setIndependentFeatures(independentFeatures)
    pnb.setDescription(description)
    pnb.setTitle(title)
    pnb.setRuntime(runtime)
    pnb.setImplementedWith(implementedWith)
    director = PretrainedNeedsDirector()
    return director.construct(pnb)


def create_molecular_req(model, title, description, type):
    pnb = PretrainedNeedsBuilder()
    independentFeatures = []
    independentFeatures.append("Smiles")
    if model.external_feats:
        independentFeatures.append(model.external_feats)
        independentFeatures = [item for sublist in independentFeatures for item in (sublist if isinstance(sublist, list) else [sublist])]
    if isinstance(model.Y, list):
        dependendFeatures = model.Y
    else:
        dependendFeatures = [model.Y]
    dependendFeatures.extend(['AD', 'Probabilities'])
    pnb.setRawModel(model)
    pnb.setJaqpotPyVersion(jaqpotpy.__version__)
    pnb.setJaqpotPyDockerVersion(config.jaqpotpy_docker)
    pnb.setLibraries(model.library)
    pnb.setVersions(model.version)
    pnb.setDependendFeatures(dependendFeatures)
    pnb.setIndependentFeatures(independentFeatures)
    pnb.setDescription(description)
    pnb.setTitle(title)
    pnb.setType(type)
    pnb.setRuntime('jaqpotpy')
    director = PretrainedNeedsDirector()
    return director.construct(pnb)


def create_doa(inv_m, a, modelid):
    doaB = DoaBuilder()
    doaB.set_aValue(a)
    doaB.set_doa_matrix(inv_m)
    doaB.set_modelid(modelid)
    director = DoaDirector()
    return director.construct(doaB)



"""
Basic molecular features.
"""
from typing import Any
import numpy as np
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
from rdkit.Chem import Descriptors
import pickle
# import dill as pickle

RDKitMol = Any


class Wrapper(object):
    def __init__(self, method_name, module_name):
        self.method_name = method_name
        self.module_name = module_name

    def __call__(self, *args, **kwargs):
        method = __import__(self.module_name, globals(), locals(), [self.method_name,])
        return method(*args, **kwargs)


class RDKitDescriptors(MolecularFeaturizer):
    """RDKit descriptors.
      This class computes a list of chemical descriptors like
      molecular weight, number of valence electrons, maximum and
      minimum partial charge, etc using RDKit.
      Attributes
      ----------
      descriptors: List[str]
        List of RDKit descriptor names used in this class.
      Note
      ----
      This class requires RDKit to be installed.
      Examples
      --------
      >>> import jaqpotpy as jt
      >>> smiles = ['CC(=O)OC1=CC=CC=C1C(=O)O']
      >>> featurizer = dc.feat.RDKitDescriptors()
      >>> features = featurizer.featurize(smiles)
      >>> type(features[0])
      <class 'numpy.ndarray'>
      >>> features[0].shape
      (208,)
      """

    def pick(self):
        with open("test.picl", "wb") as p:
            pickle.dump(self, p)
            # pickle.dump(self, p)

    @classmethod
    def unpick(cls, obj):
        with open(obj, "rb") as p:
            return pickle.load(p)

    @property
    def __name__(self):
        return 'RDKitDescriptors'

    def __init__(self, use_fragment=True, ipc_avg=True):
        """Initialize this featurizer.
        Parameters
        ----------
        use_fragment: bool, optional (default True)
          If True, the return value includes the fragment binary descriptors like 'fr_XXX'.
        ipc_avg: bool, optional (default True)
          If True, the IPC descriptor calculates with avg=True option.
          Please see this issue: https://github.com/rdkit/rdkit/issues/1527.
        """
        self.use_fragment = use_fragment
        self.ipc_avg = ipc_avg
        self.descriptors = []
        self.descList = []

    def __getitem__(self):
        return self

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate RDKit descriptors.
        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
          RDKit Mol object
        Returns
        -------
        np.ndarray
          1D array of RDKit descriptors for `mol`.
          The length is `len(self.descriptors)`.
        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        # initialize
        if len(self.descList) == 0:
            try:
                for descriptor, function in Descriptors.descList:
                    if self.use_fragment is False and descriptor.startswith('fr_'):
                        continue
                    self.descriptors.append(descriptor)
                    self.descList.append((descriptor, function))
            except ModuleNotFoundError:
                raise ImportError("This class requires RDKit to be installed.")

        # check initialization
        assert len(self.descriptors) == len(self.descList)

        features = []
        for desc_name, function in self.descList:
            if desc_name == 'Ipc' and self.ipc_avg:
                feature = function(datapoint, avg=True)
            else:
                feature = function(datapoint)
            features.append(feature)
        return np.asarray(features)

    def _get_column_names(self, **kwargs) -> list:
        descriptors = []
        try:
            for descriptor, function in Descriptors.descList:
                if self.use_fragment is False and descriptor.startswith('fr_'):
                    continue
                descriptors.append(descriptor)
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")
        return descriptors

    def _featurize_dataframe(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate RDKit descriptors.
        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
          RDKit Mol object
        Returns
        -------
        np.ndarray
          1D array of RDKit descriptors for `mol`.
          The length is `len(self.descriptors)`.
        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        # initialize
        if len(self.descList) == 0:
            try:
                for descriptor, function in Descriptors.descList:
                    if self.use_fragment is False and descriptor.startswith('fr_'):
                        continue
                    self.descriptors.append(descriptor)
                    self.descList.append((descriptor, function))
            except ModuleNotFoundError:
                raise ImportError("This class requires RDKit to be installed.")

        # check initialization
        assert len(self.descriptors) == len(self.descList)

        features = []
        for desc_name, function in self.descList:
            if desc_name == 'Ipc' and self.ipc_avg:
                feature = function(datapoint, avg=True)
            else:
                feature = function(datapoint)
            features.append(feature)
        return np.asarray(features)



import inspect
import logging
import numpy as np
import pandas as pd
from typing import Any, Iterable, Union, List, Generator
import os
from jaqpotpy.entities.material_models import Atoms

logger = logging.getLogger(__name__)
_print_threshold = 10

class Parser(object):
  """
  Abstract class for parsing molecules or materials files into serializable objects.
  This class is abstract and cannot be invoked directly. You'll
  likely only interact with this class if you're a developer. In
  that case, you might want to make a child class which
  implements the `_parse` (or the `_parse_dataframe`) method for calculating parsing for
  parsing files representing materials or molecules.
  """

  def __init__(self, path: str, file_ext: Union[str, List[str]]):
    self.path = path
    if self.path[-1] == '/':
      self.path = self.path[:-1]

    self.file_ext = file_ext
    if isinstance(self.file_ext, str):
      self.file_ext = [self.file_ext]

    self.files_ = []

  def parse(self) -> Generator:
    """
    Parse files representing materials or molecules.

    Parameters
    ----------
    path: str
      Either the path of a certain file or a path of a folder containing
      files that will be parsed

    file_ext: str | List[str]
      The extention of the files that will be parsed.

    log_every_n: int, default 1000
      Logs featurization progress every `log_every_n` steps.

    Returns
    -------
    generator
      A generator with the parsed objects of the files.
    """

    try:
      if self.path.split('.')[-1].lower() not in self.file_ext:
        try:
          file = [self.path+'/'+item for item in os.listdir(self.path) if item.split('.')[-1].lower() in self.file_ext]
        except:
          raise ValueError('Invalid file type. Expected {} but found {} instead'.format(self.file_ext, self.path.split('.')[-1]))
      else:
        file = [self.path]
    except:
      try:
        file = [self.path+'/'+item for item in os.listdir(self.path) if item.split('.')[-1].lower() in self.file_ext]
      except:
        raise ValueError("Invalid file type. Expected {} but didn't find any".format(self.file_ext))

    # Open the file and read it in as a list of rows
    for i, item in enumerate(file):
      try:
        obj = self._parse(item)
        if isinstance(obj, list):
          for i in obj:
            yield i
        else:
           yield obj
      except Exception as e:
        logger.warning("Failed to parse file {}.".format(item), e)


  def parse_dataframe(self) -> Union[List[pd.DataFrame], pd.DataFrame]:
    """
    Parse files of materials or molecules in a pandas dataframe.

    Parameters
    ----------
    path: str
      Either the path of a certain file or a path of a folder containing
      files that will be parsed

    file_ext: str | List[str]
      The extention of the files that will be parsed.

    log_every_n: int, default 1000
      Logs featurization progress every `log_every_n` steps.

    Returns
    -------
    pd.DataFrame
      A pandas dataframe of the files.
    """

    files = self.parse()

    df = pd.DataFrame()
    df_list = []
    for i, file in enumerate(files):
      resp = self._parse_dataframe(file, self.files_[i])
      try:
        df = pd.concat([df, resp]).reset_index(drop=True)
      except:
        if len(df_list) == 0:
          df_list = [pd.concat([df, item]).reset_index(drop=True) for item in resp]
        else:
          for j in range(len(df_list)):
            df_list[j] = pd.concat([df_list[j], resp[j]]).reset_index(drop=True)

    if len(df_list)>0:
      return df_list
    return df

  def __call__(self, path: str, file_ext: Union[str, List[str]], **kwargs):
    """Calculate features for datapoints.
    `**kwargs` will get passed directly to `Featurizer.featurize`
    Parameters
    ----------
    datapoints: Iterable[Any]
      Any blob of data you like. Subclasss should instantiate this.
    """
    return self.parse()

  def _parse(self, path: str):
    """Calculate features for a single datapoint.
    Parameters
    ----------
    datapoint: Any
      Any blob of data you like. Subclass should instantiate this.
    """
    raise NotImplementedError('Parser is not defined.')

  def _parse_dataframe(self, file: Any, filename: str) -> Union[List[pd.DataFrame], pd.DataFrame]:
    """Calculate features for a single datapoint.
    Parameters
    ----------
    datapoint: Any
      Any blob of data you like. Subclass should instantiate this.
    """
    raise NotImplementedError('Parser is not defined.')


  def __repr__(self) -> str:
    """Convert self to repr representation.
    Returns
    -------
    str
      The string represents the class.
    Examples
    --------
    >>> import jaqpotpy as jt
    >>> jt.parsers.PdbParser('pdb_file.pdb')
    PdbParser[pdb='pdb_file.pdb', file_ext='pdb']
    """
    args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
    args_names = [arg for arg in args_spec.args if arg != 'self']
    args_info = ''
    for arg_name in args_names:
      value = self.__dict__[arg_name]
      # for str
      if isinstance(value, str):
        value = "'" + value + "'"
      # for list
      if isinstance(value, list):
        threshold = get_print_threshold()
        value = np.array2string(np.array(value), threshold=threshold)
      args_info += arg_name + '=' + str(value) + ', '
    return self.__class__.__name__ + '[' + args_info[:-2] + ']'

  def __str__(self) -> str:
    """Convert self to str representation.
    Returns
    -------
    str
      The string represents the class.
    Examples
    --------
    >>> import jaqpotpy as jt
    >>> jt.parsers.PdbParser('pdb_file.pdb')
    PdbParser[pdb='pdb_file.pdb', file_ext='pdb']
    """
    args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
    args_names = [arg for arg in args_spec.args if arg != 'self']
    args_num = len(args_names)
    args_default_values = [None for _ in range(args_num)]
    if args_spec.defaults is not None:
      defaults = list(args_spec.defaults)
      args_default_values[-len(defaults):] = defaults

    override_args_info = ''
    for arg_name, default in zip(args_names, args_default_values):
      if arg_name in self.__dict__:
        arg_value = self.__dict__[arg_name]
        # validation
        # skip list
        if isinstance(arg_value, list):
          continue
        if isinstance(arg_value, str):
          # skip path string
          if "\\/." in arg_value or "/" in arg_value or '.' in arg_value:
            continue
        # main logic
        if default != arg_value:
          override_args_info += '_' + arg_name + '_' + str(arg_value)
    return self.__class__.__name__ + override_args_info

  def to_xyz(self):
    xyzs = []
    files = self.parse()
    for file in files:
      atoms:Atoms = file.get_atoms()
      xyz = str(len(atoms.elements)) + '\ntemp xyz\n'
      for i in range(len(atoms.elements)):
        xyz += atoms.elements[i] + '\t' + str(atoms.coordinates[i][0]) + '\t' + str(atoms.coordinates[i][1]) + '\t' + str(atoms.coordinates[i][2]) + '\n'

      xyzs.append(xyz)

    return xyzs


def get_print_threshold() -> int:
  """Return the printing threshold for datasets.
  The print threshold is the number of elements from ids/tasks to
  print when printing representations of `Dataset` objects.
  Returns
  ----------
  threshold: int
    Number of elements that will be printed
  """
  return _print_threshold


import torch
from jaqpotpy.cfg import config
import torch.nn.functional as F
import math


class CNN(torch.nn.Module):
    def __init__(self, layers: torch.nn.ModuleList, out_channels, input_size, dropout=0):
        super(CNN, self).__init__()

        torch.manual_seed(config.global_seed)

        self.dropout = dropout
        self.layers = layers
        self.input_size = input_size

        # fully connected layer
        self.out = torch.nn.Linear(self.linear_input_size(), out_channels)

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        # flatten the output to (batch_size, 32 * 7 * 7)
        x = x.view(x.size(0), -1)
        output = self.out(x)
        return output  # return x for visualization

    # after obtaining the size in above method, we call it and multiply all elements of the returned size.
    def linear_input_size(self):
        dummy = torch.rand(1, self.layers[0].__dict__['in_channels'], self.input_size[0],
                           self.input_size[1])  # image size: 64x32
        return math.prod(self.layers(dummy).size())



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='jaqpotpy',
      version='6.0.0',
      description='Standardizing molecular modeling',
      long_description=long_description,
      long_description_content_type='text/markdown',

      url='https://github.com/ntua-unit-of-control-and-informatics/jaqpotpy',
      author='Unit of Process Control and Informatics | National Technical University of Athens',
      author_email='upci.ntua@gmail.com',
      license='MIT License',
      packages=find_packages(exclude=["*.tests"]),
      package_data={'jaqpotpy': ['data/*.gz']},
      install_requires=[
            'pandas==2.2.2',
            'pyjwt==2.8.0',
            'simplejson==3.19.2',
            'pydotplus==2.0.2',
            'requests==2.32.2',
            'pydantic==2.7.1',
            'rdkit-pypi==2022.9.5',
            'mordred==1.2.0',
            'scikit-learn==1.5.0',
            'tqdm==4.66.4',
            'kennard-stone==2.2.1',
            'mendeleev==0.16.2',
            'pymatgen==2024.5.1',
            'skl2onnx==1.16.0',
            'onnxruntime==1.18.0',
            'torch==2.3.0',
            'torch-geometric==2.3.1',
            'torchvision==0.18.0',
      ],
      zip_safe=False)



import datetime
import os
import time

import rdkit

from jaqpotpy.models.generative.models import GanMoleculeGenerator, MoleculeDiscriminator
from jaqpotpy.datasets.molecular_datasets import SmilesDataset
import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from jaqpotpy.descriptors.molecular import MolGanFeaturizer, GraphMatrix
from rdkit import Chem
from jaqpotpy.helpers.logging import init_logger
from typing import Any
import numpy as np
from jaqpotpy.models.evaluator import GenerativeEvaluator
from collections import defaultdict
from rdkit import RDLogger

from jaqpotpy.models.generative.molecular_metrics import MolecularMetrics

RDLogger.DisableLog('rdApp.info')


class GanSolver(object):

    def __init__(self, generator: GanMoleculeGenerator
                 , discriminator: MoleculeDiscriminator
                 , dataset: SmilesDataset
                 , evaluator: GenerativeEvaluator
                 , g_lr
                 , d_lr
                 , lamda_gradient_penalty: float = 20.0
                 , la: float = 0.5
                 , n_critic: int = 3
                 , weight_decay: float = 5e-4
                 , optimizer: str = "Adam"
                 , batch_size: int = 32
                 , epochs: int = 12
                 , val_at: int = 10
                 , post_method: str = "softmax"
                 , metric='validity,qed'
                 , device=torch.device('cuda' if torch.cuda.is_available() else 'cpu')
                 , path="./"
                 , task: str = "Generate / Reinforce"
                 # , OptimizationModel: MolecularModel
                 ):
        self.start_time = time.time()
        self.generator = generator
        self.discriminator = discriminator
        self.value_network = discriminator
        # self.generator = GanMoleculeGenerator([128, 256, 524], 8, 42, 5, 12, 0.5)
        # self.discriminator = MoleculeDiscriminator([[128, 64], 128, [128, 64]], 12, 5 - 1, 0.5)
        # self.value_network = MoleculeDiscriminator([[128, 64], 128, [128, 64]], 1, 5 - 1, 0.5)
        self.post_method = post_method
        self.dataset = dataset
        self.batch_size = batch_size
        self.device = device

        self.metric = metric

        self.evaluator = evaluator

        self.la = la
        self.g_lr = g_lr
        self.d_lr = d_lr
        self.la_gp = lamda_gradient_penalty
        self.weight_decay = weight_decay

        self.optimizer = optimizer
        self.g_optim = None
        self.d_optim = None
        self.v_optim = None

        self.logger = init_logger(__name__, testing_mode=False, output_log_file=False)

        self.loader_params = {'batch_size': self.batch_size, 'shuffle': True, 'num_workers': 0}
        self.data_loader = DataLoader(self.dataset, **self.loader_params)

        self.val_after = val_at
        self.epochs = epochs
        self.num_steps = (len(self.dataset) // self.batch_size)

        self.image_dir_path = path

        if self.la > 0:
            self.n_critic = n_critic
        else:
            self.n_critic = 1

        self.BOND_DIM = self.dataset.featurizer.BOND_DIM
        self.MAX_ATOMS = self.dataset.featurizer.MAX_ATOMS
        self.build_env()

    def build_env(self):

        if self.optimizer == "Adam":
            self.g_optim = torch.optim.Adam(self.generator.parameters(), lr=self.g_lr, weight_decay=self.weight_decay)
            self.d_optim = torch.optim.Adam(self.discriminator.parameters(), lr=self.d_lr,
                                            weight_decay=self.weight_decay)
            self.v_optim = torch.optim.Adam(self.value_network.parameters(), lr=self.d_lr,
                                            weight_decay=self.weight_decay)
        if self.optimizer == "AdamW":
            self.g_optim = torch.optim.AdamW(self.generator.parameters(), lr=self.g_lr, weight_decay=self.weight_decay)
            self.d_optim = torch.optim.AdamW(self.discriminator.parameters(), lr=self.d_lr,
                                             weight_decay=self.weight_decay)
            self.v_optim = torch.optim.AdamW(self.value_network.parameters(), lr=self.d_lr,
                                             weight_decay=self.weight_decay)
        if self.optimizer == "SGD":
            self.g_optim = torch.optim.SGD(self.generator.parameters(), lr=self.g_lr, weight_decay=self.weight_decay)
            self.d_optim = torch.optim.SGD(self.discriminator.parameters(), lr=self.d_lr,
                                           weight_decay=self.weight_decay)
            self.v_optim = torch.optim.SGD(self.value_network.parameters(), lr=self.d_lr,
                                           weight_decay=self.weight_decay)
        if self.optimizer == "Adadelta":
            self.g_optim = torch.optim.Adadelta(self.generator.parameters(), lr=self.g_lr,
                                                weight_decay=self.weight_decay)
            self.d_optim = torch.optim.Adadelta(self.discriminator.parameters(), lr=self.d_lr,
                                                weight_decay=self.weight_decay)
            self.v_optim = torch.optim.Adadelta(self.value_network.parameters(), lr=self.d_lr,
                                                weight_decay=self.weight_decay)
        if self.optimizer == "Adagrad":
            self.g_optim = torch.optim.Adagrad(self.generator.parameters(), lr=self.g_lr,
                                               weight_decay=self.weight_decay)
            self.d_optim = torch.optim.Adagrad(self.discriminator.parameters(), lr=self.d_lr,
                                               weight_decay=self.weight_decay)
            self.v_optim = torch.optim.Adagrad(self.value_network.parameters(), lr=self.d_lr,
                                               weight_decay=self.weight_decay)
        if self.optimizer == "Adamax":
            self.g_optim = torch.optim.Adamax(self.generator.parameters(), lr=self.g_lr, weight_decay=self.weight_decay)
            self.d_optim = torch.optim.Adamax(self.discriminator.parameters(), lr=self.d_lr,
                                              weight_decay=self.weight_decay)
            self.v_optim = torch.optim.Adamax(self.value_network.parameters(), lr=self.d_lr,
                                              weight_decay=self.weight_decay)
        if self.optimizer == "RMSprop":
            self.g_optim = torch.optim.RMSprop(self.generator.parameters(), lr=self.g_lr,
                                               weight_decay=self.weight_decay)
            self.d_optim = torch.optim.RMSprop(self.discriminator.parameters(), lr=self.d_lr,
                                               weight_decay=self.weight_decay)
            self.v_optim = torch.optim.RMSprop(self.value_network.parameters(), lr=self.d_lr,
                                               weight_decay=self.weight_decay)

        self.discriminator.to(self.device)
        self.value_network.to(self.device)
        self.generator.to(self.device)

    def reset_grad(self):
        """Reset the gradient buffers."""
        self.g_optim.zero_grad()
        self.d_optim.zero_grad()
        self.v_optim.zero_grad()

    def fit(self):
        the_step = self.num_steps
        for epoch_i in range(self.epochs):
            for a_step, i in enumerate(self.data_loader):
                nod_fs = torch.split(i[2][0], 1)
                adc_mat = torch.split(i[2][1], 1)
                mols = []
                mol_strings = []
                for ind, n in enumerate(nod_fs):
                    g2 = GraphMatrix(torch.squeeze(adc_mat[ind]).cpu().detach().numpy(),
                                     torch.squeeze(n).cpu().detach().numpy())
                    m = self.dataset.featurizer._defeaturize(g2, sanitize=False, cleanup=False)
                    mol_strings.append(Chem.MolToSmiles(m))
                    mols.append(m)

                if epoch_i < 0:
                    cur_la = 0
                else:
                    cur_la = self.la

                cur_step = self.num_steps * epoch_i + a_step

                # =================================================================================== #
                #                             2. Train the discriminator                              #
                # =================================================================================== #

                a_tensor = torch.permute(i[2][1], (0, 2, 3, 1))
                x_tensor = i[2][0]

                logits_real, features_real = self.discriminator(a_tensor, None, x_tensor)

                z = self.sample_z(i[2][0].size(dim=0))

                edges_logits, nodes_logits = self.generator(z)
                (edges_hat, nodes_hat) = self.postprocess((edges_logits, nodes_logits), self.post_method)

                logits_fake, features_fake = self.discriminator(edges_hat, None, nodes_hat)

                # Compute losses for gradient penalty.
                eps = torch.rand(logits_real.size(0), 1, 1, 1).to(self.device)
                x_int0 = (eps * a_tensor + (1. - eps) * edges_hat).requires_grad_(True)
                x_int1 = (eps.squeeze(-1) * x_tensor + (1. - eps.squeeze(-1)) * nodes_hat).requires_grad_(True)
                grad0, grad1 = self.discriminator(x_int0, None, x_int1)
                grad_penalty = self.gradient_penalty(grad0, x_int0) + self.gradient_penalty(grad1, x_int1)

                d_loss_real = torch.mean(logits_real)
                d_loss_fake = torch.mean(logits_fake)
                loss_D = -d_loss_real + d_loss_fake + self.la_gp * grad_penalty

                if cur_step % self.n_critic != 0 and cur_la > 0:
                    self.reset_grad()
                    # self.d_optim.zero_grad()
                    loss_D.backward(inputs=list(self.discriminator.parameters()))
                    self.d_optim.step()

                # =================================================================================== #
                #                               3. Train the generator                                #
                # =================================================================================== #

                edges_logits, nodes_logits = self.generator(z)
                (edges_hat, nodes_hat) = self.postprocess((edges_logits, nodes_logits), self.post_method)
                logits_fake, features_fake = self.discriminator(edges_hat, None, nodes_hat)

                # logits_fake, features_fake = out_a_hat, logits_hat
                value_logit_real, _ = self.value_network(a_tensor, None, x_tensor, torch.sigmoid)
                value_logit_fake, _ = self.value_network(edges_hat, None, nodes_hat, torch.sigmoid)

                mols_hat = []

                nod_fs = torch.split(nodes_hat, 1)
                adc_mat = torch.split(edges_hat, 1)
                for ind, n in enumerate(nod_fs):

                    adjacency = torch.argmax(adc_mat[ind], dim=3)
                    adjacency = torch.nn.functional.one_hot(adjacency, num_classes=self.BOND_DIM)
                    adjacency = torch.squeeze(adjacency)
                    adjacency = torch.permute(adjacency, (2, 0, 1))

                    nodes = torch.argmax(n, dim=2)
                    nodes = torch.nn.functional.one_hot(nodes, num_classes=self.MAX_ATOMS)

                    g2 = GraphMatrix(adjacency.detach().numpy(),
                                     torch.squeeze(nodes).cpu().detach().numpy())

                    t = self.dataset.featurizer.defeaturize(g2)
                    if t.size == 0:
                        mols_hat.append(None)
                    else:
                        if t[0] is not None:
                            print(Chem.MolToSmiles(t[0]))
                        mols_hat.append(t[0])

                # Real Reward
                reward_r = torch.from_numpy(self.evaluator.get_reward(mols)).to(self.device)
                # Fake Reward
                reward_f = torch.from_numpy(self.evaluator.get_reward(mols_hat)).to(self.device)

                # # Real Reward (OLD MOLECULAR METRICS)
                # reward_r = torch.from_numpy(self.reward(mols)).to(self.device)
                # # Fake Reward (OLD MOLECULAR METRICS)
                # reward_f = torch.from_numpy(self.reward(mols_hat)).to(self.device)

                # Losses Update
                loss_G = -logits_fake
                # Original TF loss_V. Here we use absolute values instead of the squared one.
                # loss_V = (value_logit_real - reward_r) ** 2 + (value_logit_fake - reward_f) ** 2
                loss_V = torch.abs(value_logit_real - reward_r) + torch.abs(value_logit_fake - reward_f)
                loss_RL = -value_logit_fake

                loss_G = torch.mean(loss_G)
                loss_V = torch.mean(loss_V)
                loss_RL = torch.mean(loss_RL)

                alpha = torch.abs(loss_G.detach() / loss_RL.detach()).detach()
                # alpha = torch.abs(loss_G / loss_RL)
                train_step_G = cur_la * loss_G + (1 - cur_la) * alpha * loss_RL
                #
                train_step_V = loss_V

                # alpha = torch.abs(loss_G.detach() / loss_RL.detach()).detach()
                # train_step_G = cur_la * loss_G.cpu().detach().numpy() + (1 - cur_la) * alpha.cpu().detach().numpy() * loss_RL.cpu().detach().numpy()
                # train_step_G = torch.from_numpy(np.full(1, train_step_G))
                # train_step_G.requires_grad = True
                #
                # train_step_V = loss_V.cpu().detach().numpy()
                # train_step_V = torch.from_numpy(np.full(1, train_step_V))
                # train_step_V.requires_grad = True

                self.logger.info("Reward from reals for epoch "
                      + str(epoch_i) + " step " + str(cur_step - self.num_steps * epoch_i)
                      + ": " + str(np.mean(reward_r.cpu().detach().numpy())))
                self.logger.info("Reward from fakes for epoch "
                      + str(epoch_i) + " step " + str(cur_step - self.num_steps * epoch_i)
                      + ": " + str(np.mean(reward_f.cpu().detach().numpy())))
                self.logger.info("Generator loss for epoch "
                                 + str(epoch_i) + " step " + str(cur_step - self.num_steps * epoch_i)
                                 + ": " + str(loss_G.cpu().detach().numpy()))
                self.logger.info("Discriminator loss for epoch "
                                 + str(epoch_i) + " step " + str(cur_step - self.num_steps * epoch_i)
                                 + ": " + str(loss_V.cpu().detach().numpy()))
                self.logger.info("Value loss for epoch "
                                 + str(epoch_i) + " step " + str(cur_step - self.num_steps * epoch_i)
                                 + ": " + str(loss_RL.cpu().detach().numpy()))

                self.reset_grad()

                if cur_step % self.n_critic == 0:
                    # self.g_optim.zero_grad()
                    train_step_G.backward(retain_graph=True, inputs=list(self.generator.parameters()))
                    self.g_optim.step()

                # Optimise value network.
                if cur_step % self.n_critic == 0:
                    # self.v_optim.zero_grad()
                    train_step_V.backward(inputs=list(self.value_network.parameters()))
                    self.v_optim.step()

                losses = defaultdict(list)
                scores = defaultdict(list)

                if epoch_i > self.val_after and cur_step - self.num_steps * epoch_i == 0:
                    self.logger.info("VALIDATING")
                    mols = mols_hat
                    self.all_scores(mols, self.evaluator.dataset, epoch_i=epoch_i, norm=True)
                    # self.all_scores(mols, self.evaluator.dataset, epoch_i=epoch_i, norm=True)  # 'mols' is output of Fake Reward
            pass

    def sample_z(self, batch_size):
        return torch.Tensor(np.random.normal(0, 1, size=(batch_size, 8)))

    def get_mols(self, data: Any):
        return ""

    def postprocess(self, inputs, method, temperature=1.):

        def listify(x):
            return x if type(x) == list or type(x) == tuple else [x]

        def delistify(x):
            return x if len(x) > 1 else x[0]

        if method == 'soft_gumbel':
            softmax = [F.gumbel_softmax(e_logits.contiguous().view(-1, e_logits.size(-1))
                                        / temperature, hard=False).view(e_logits.size())
                       for e_logits in listify(inputs)]
        elif method == 'hard_gumbel':
            softmax = [F.gumbel_softmax(e_logits.contiguous().view(-1, e_logits.size(-1))
                                        / temperature, hard=True).view(e_logits.size())
                       for e_logits in listify(inputs)]
        else:
            softmax = [F.softmax(e_logits / temperature, -1)
                       for e_logits in listify(inputs)]

        return [delistify(e) for e in (softmax)]

    def gradient_penalty(self, y, x):
        """Compute gradient penalty: (L2_norm(dy/dx) - 1)**2."""
        weight = torch.ones(y.size()).to(self.device)
        dydx = torch.autograd.grad(outputs=y,
                                   inputs=x,
                                   grad_outputs=weight,
                                   retain_graph=True,
                                   create_graph=True,
                                   only_inputs=True)[0]

        dydx = dydx.view(dydx.size(0), -1)
        dydx_l2norm = torch.sqrt(torch.sum(dydx ** 2, dim=1))
        return torch.mean((dydx_l2norm - 1) ** 2)

    def print_network(self, model, name, log=None):
        """Print out the network information."""
        num_params = 0
        for p in model.parameters():
            num_params += p.numel()
        if log is not None:
            self.logger.log(model)
            self.logger.log(name)
            self.logger.log("The number of parameters: {}".format(num_params))
        else:
            print(model)
            print(name)
            print("The number of parameters: {}".format(num_params))

    def all_scores_(self, mols, data, epoch_i, norm=False, reconstruction=False):
        m0 = {k: list(filter(lambda e: e is not None, v)) for k, v in {
                'NP': MolecularMetrics.natural_product_scores(mols, norm=norm),
                'QED': MolecularMetrics.quantitative_estimation_druglikeness_scores(mols),
                'Solute': MolecularMetrics.water_octanol_partition_coefficient_scores(mols, norm=norm),
                # 'SA': MolecularMetrics.synthetic_accessibility_score_scores(mols, norm=norm),
                # 'diverse': MolecularMetrics.diversity_scores(mols, data),
                'drugcand': MolecularMetrics.drugcandidate_scores(mols, data)}.items()}

        m1 = {'valid': MolecularMetrics.valid_total_score(mols) * 100,
                  'unique': MolecularMetrics.unique_total_score(mols) * 100,
                  'novel': MolecularMetrics.novel_total_score(mols, data) * 100}
        losses = defaultdict(list)
        scores = defaultdict(list)
        for k, v in m1.items():
            scores[k].append(v)
        for k, v in m0.items():
            scores[k].append(np.array(v)[np.nonzero(v)].mean())

            # Saving molecule images.
        mol_f_name = os.path.join(self.image_dir_path, 'mol-{}.png'.format(epoch_i))
        save_mol_img(mols, mol_f_name)

        # Print out training information.
        et = time.time() - self.start_time
        et = str(datetime.timedelta(seconds=et))[:-7]
        log = "Elapsed [{}], Iteration [{}/{}]:".format(et, epoch_i + 1, self.epochs)

        is_first = True
        for tag, value in losses.items():
            if is_first:
                log += "\n{}: {:.2f}".format(tag, np.mean(value))
                is_first = False
            else:
                log += ", {}: {:.2f}".format(tag, np.mean(value))
        is_first = True
        for tag, value in scores.items():
            if is_first:
                log += "\n{}: {:.2f}".format(tag, np.mean(value))
                is_first = False
            else:
                log += ", {}: {:.2f}".format(tag, np.mean(value))
        print(log)
        if self.logger is not None:
            self.logger.info(log)

    def all_scores(self, mols, data, epoch_i, norm=False, reconstruction=False):
        et = time.time() - self.start_time
        et = str(datetime.timedelta(seconds=et))[:-7]
        log = "Elapsed [{}], Iteration [{}/{}]:".format(et, epoch_i + 1, self.epochs)
        for key in self.evaluator.eval_functions.keys():
            function_name = key
            f = self.evaluator.eval_functions.get(key)
            try:
                score = f(mols)
                log += ", {}: {:.2f}".format(function_name, np.mean(score))
            except TypeError as e:
                score = f(mols, data)
                log += ", {}: {:.2f}".format(function_name, np.mean(score))
        if self.logger is not None:
            self.logger.info(log)
        else:
            print(log)
        mol_f_name = os.path.join(self.image_dir_path, 'mol-{}.png'.format(epoch_i))
        save_mol_img(mols, mol_f_name)

    def reward(self, mols):
        rr = 1.
        for m in ('logp,sas,qed,unique' if self.metric == 'all' else self.metric).split(','):

            if m == 'np':
                rr *= MolecularMetrics.natural_product_scores(mols, norm=True)
            elif m == 'logp':
                rr *= MolecularMetrics.water_octanol_partition_coefficient_scores(mols, norm=True)
            elif m == 'sas':
                rr *= MolecularMetrics.synthetic_accessibility_score_scores(mols, norm=True)
            elif m == 'qed':
                rr *= MolecularMetrics.quantitative_estimation_druglikeness_scores(mols, norm=True)
            elif m == 'novelty':
                rr *= MolecularMetrics.novel_scores(mols, self.evaluator.dataset)
            elif m == 'dc':
                rr *= MolecularMetrics.drugcandidate_scores(mols, self.evaluator.dataset)
            elif m == 'unique':
                rr *= MolecularMetrics.unique_scores(mols)
            elif m == 'diversity':
                rr *= MolecularMetrics.diversity_scores(mols, self.evaluator.dataset)
            elif m == 'validity':
                rr *= MolecularMetrics.valid_scores(mols)
            else:
                raise RuntimeError('{} is not defined as a metric'.format(m))

        return rr.reshape(-1, 1)


import datetime
import string
import random

def save_mol_img(mols, f_name='tmp.png', is_test=False):
    orig_f_name = f_name
    for a_mol in mols:
        try:
            if Chem.MolToSmiles(a_mol) is not None:
                print('Generating molecule')

                if is_test:
                    f_name = orig_f_name
                    f_split = f_name.split('.')
                    f_split[-1] = random_string() + '.' + f_split[-1]
                    f_name = ''.join(f_split)

                rdkit.Chem.Draw.MolToFile(a_mol, f_name)
                break

                # if not is_test:
                #     break
        except:
            continue



def random_string(string_len=3):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(string_len))


# from tornado import gen, httpclient
# from tornado.httputil import HTTPHeaders
# from jaqpotpy.mappers import decode
import requests
import json


chempot_path = "chempot"

def predict(baseurl, api_key, modelid, smiles, descriptors, doa, logger):
    
    uri = baseurl + chempot_path + "/" 

    h = {"Content-type": "application/json",
         "Accept": "application/json",
         'Authorization': "Bearer " + api_key}
    
    data = {
      "smiles": smiles,
      "modelId": modelid,
      "withDoa": doa,
      "descriptors": descriptors
    }
    
    try:
        r = requests.post(uri, data=json.dumps(data), headers=h)
        if r.status_code < 300:
            return r.json()
        else:
            logger.error("Error with code " + str(r.status_code))
            logger.error(r.json())
    except Exception as e:
        logger.error("Error http: " + str(e))



import numpy as np
from typing import Union, Tuple
import joblib
import gzip
import pickle
import os
import tempfile
import tarfile
import zipfile
import logging
from urllib.request import urlretrieve
from typing import Any, Iterator, List, Optional, Tuple, Union, cast, IO


def pad_array(x: np.ndarray,
              shape: Union[Tuple, int],
              fill: float = 0.0,
              both: bool = False) -> np.ndarray:
  """
  Pad an array with a fill value.
  Parameters
  ----------
  x: np.ndarray
    A numpy array.
  shape: Tuple or int
    Desired shape. If int, all dimensions are padded to that size.
  fill: float, optional (default 0.0)
    The padded value.
  both: bool, optional (default False)
    If True, split the padding on both sides of each axis. If False,
    padding is applied to the end of each axis.
  Returns
  -------
  np.ndarray
    A padded numpy array
  """
  x = np.asarray(x)
  if not isinstance(shape, tuple):
    shape = tuple(shape for _ in range(x.ndim))
  pad = []
  for i in range(x.ndim):
    diff = shape[i] - x.shape[i]
    assert diff >= 0
    if both:
      a, b = divmod(diff, 2)
      b += a
      pad.append((a, b))
    else:
      pad.append((0, diff))
  pad = tuple(pad)  # type: ignore
  x = np.pad(x, pad, mode='constant', constant_values=fill)
  return x


def get_data_dir() -> str:
  """Get the DeepChem data directory.
  Returns
  -------
  str
    The default path to store DeepChem data. If you want to
    change this path, please set your own path to `DEEPCHEM_DATA_DIR`
    as an environment variable.
  """
  if 'DEEPCHEM_DATA_DIR' in os.environ:
    return os.environ['DEEPCHEM_DATA_DIR']
  return tempfile.gettempdir()


def download_url(url: str,
                 dest_dir: str = get_data_dir(),
                 name: Optional[str] = None):
  """Download a file to disk.
  Parameters
  ----------
  url: str
    The URL to download from
  dest_dir: str
    The directory to save the file in
  name: str
    The file name to save it as.  If omitted, it will try to extract a file name from the URL
  """
  if name is None:
    name = url
    if '?' in name:
      name = name[:name.find('?')]
    if '/' in name:
      name = name[name.rfind('/') + 1:]
  if not os.path.exists(dest_dir):
    os.makedirs(dest_dir)
  urlretrieve(url, os.path.join(dest_dir, name))


def untargz_file(file: str,
                 dest_dir: str = get_data_dir(),
                 name: Optional[str] = None):
  """Untar and unzip a .tar.gz file to disk.
  Parameters
  ----------
  file: str
    The filepath to decompress
  dest_dir: str
    The directory to save the file in
  name: str
    The file name to save it as.  If omitted, it will use the file name
  """
  if name is None:
    name = file
  tar = tarfile.open(name)
  tar.extractall(path=dest_dir)
  tar.close()


def unzip_file(file: str,
               dest_dir: str = get_data_dir(),
               name: Optional[str] = None):
  """Unzip a .zip file to disk.
  Parameters
  ----------
  file: str
    The filepath to decompress
  dest_dir: str
    The directory to save the file in
  name: str
    The directory name to unzip it to.  If omitted, it will use the file name
  """
  if name is None:
    name = file
  if dest_dir is None:
    dest_dir = os.path.join(get_data_dir, name)
  with zipfile.ZipFile(file, "r") as zip_ref:
    zip_ref.extractall(dest_dir)


"""
Pytorch Utilities.
This file contains utilities that compute Datastractures adn other for Pytorch.
"""
import numpy as np


def to_torch_graph_data_and_y(g, y):
    """
    Returns torch graph data from jaqpot graph_data and adds y

    """
    import torch
    from torch_geometric.data import Data
    torch_graph = Data(x=torch.FloatTensor(g.node_features)
                , edge_index=torch.LongTensor(g.edge_index)
                , edge_attr=g.edge_features
                , num_nodes=g.num_nodes, y=np.array([y]))
    return torch_graph


def to_torch_graph_data(g):
    """
    Returns torch graph data from jaqpot graph_data without endpoint
    """
    import torch
    from torch_geometric.data import Data
    torch_graph = Data(x=torch.FloatTensor(g.node_features)
                , edge_index=torch.LongTensor(g.edge_index)
                , edge_attr=g.edge_features
                , num_nodes=g.num_nodes)
    return torch_graph


def to_torch_graph_data_array_and_regr_y(mol_g, y):
    """
    Returns torch graph data from array of jaqpot graph_data and y
    """
    import torch
    from torch_geometric.data import Data
    datas = []
    i = 0
    for g in mol_g:
        dato = Data(x=torch.FloatTensor(g.node_features)
                    , edge_index=torch.LongTensor(g.edge_index)
                    , edge_attr=g.edge_features
                    , num_nodes=g.num_nodes, y=torch.Tensor([y[i]]))
        datas.append(dato)
        i += 1
    return datas


def to_torch_graph_data_array_and_class_y(mol_g, y):
    """
    Returns torch graph data from array of jaqpot graph_data and y
    """
    import torch
    from torch_geometric.data import Data
    datas = []
    i = 0
    for g in mol_g:
        dato = Data(x=torch.FloatTensor(g.node_features)
                    , edge_index=torch.LongTensor(g.edge_index)
                    , edge_attr=g.edge_features
                    , num_nodes=g.num_nodes, y=torch.LongTensor([y[i]]))
        datas.append(dato)
        i += 1
    return datas


def to_torch_graph_data_array(mol_g):
    """
    Returns torch graph data from array of jaqpot graph_data
    """
    import torch
    from torch_geometric.data import Data
    datas = []
    for g in mol_g:
        dato = Data(x=torch.FloatTensor(g.node_features)
                    , edge_index=torch.LongTensor(g.edge_index)
                    , edge_attr=g.edge_features
                    , num_nodes=g.num_nodes)
        datas.append(dato)
    return datas



from typing import List
from jaqpotpy.parsers.base_classes import Parser
from jaqpotpy.entities.material_models import (
    Atoms, Sdf, SdfHeader, SdfCTable, SdfBond, SdfCountLine
)
import pandas as pd
from pydantic import parse_obj_as


def _clean_key(word) -> str:
    return ''.join(i for i in word if i.isalnum())


class MolParser(Parser):
    """
    MolParser.
    This class parses a mol or sdf file (or all mol or sdf files in a folder) into a Sdf structure.
    For more information on the structure see jaqpotpy.models.material_models.Sdf

    Attributes
    ----------
    files_: str | List[str]
        File names of the pdb files that were parsed.

    Note
    ----

    Examples
    --------
    >>> import jaqpotpy as jt
    >>> mol_file = './caffeine.mol'
    >>> parser = jt.parsers.mol_parser.mol_parser.MolParser()
    >>> mol = parser.parse(mol_file)
    >>> type(mol)
    generator
    >>> parsed = next(mol)
    >>> type(parsed)
    jaqpotpy.models.material_models.Sdf
    """

    @property
    def __name__(self):
        return 'MolParser'


    def __getitem__(self):
        return self

    def _parse(self, path:str) -> Sdf:

        """
        Parse sdf and mol files.

        Parameters
        ----------
        path: str
          Either the path of a certain file or a path of a folder containing
          files that will be parsed

        Returns
        -------
        jaqpotpy.models.material_models.Sdf
          A Sdf object
        """
        # Initialize variables
        sdf_dict: Sdf = Sdf(header=SdfHeader(title='', timestamp='', comment=''),
                            con_table=SdfCTable(
                                count_line=SdfCountLine(n_atoms=0, n_bonds=0, n_atom_list=0, chiral_flag=0, n_stext_ents=0, n_add_props=0, version=''),
                                atoms=Atoms(elements=[], coordinates=[], extraInfo=[]),
                                bonds=[SdfBond(first_atom=0, second_atom=0, bond_type=0, bond_stereo=0, bond_topology=0, reacting_center_status=0)],
                                properties=[]),
                            extraInfo=[])

        head: SdfHeader = SdfHeader(title='', timestamp='', comment='')
        ctable: SdfCTable = SdfCTable(
            count_line=SdfCountLine(n_atoms=0, n_bonds=0, n_atom_list=0, chiral_flag=0, n_stext_ents=0, n_add_props=0, version=''),
            atoms=Atoms(elements=[], coordinates=[], extraInfo=[]),
            bonds=[SdfBond(first_atom=0, second_atom=0, bond_type=0, bond_stereo=0, bond_topology=0, reacting_center_status=0)],
            properties=[])
        records = []

        # Open the file and read it in as a list of rows
        with open(path) as f:
            sdf = f.read().splitlines()

        # Iterate through the file
        line_cnt = 0
        extraProp = ''
        cont = True

        for row in sdf:
            if line_cnt == 0:
                head.title = row
            elif line_cnt == 1:
                head.timestamp = row
            elif line_cnt == 2:
                head.comment = row
            elif line_cnt == 3:
                curr_list = row.split()
                try:
                    ctable.count_line.n_atoms = int(curr_list[0])
                    ctable.count_line.n_bonds = int(curr_list[1])
                    ctable.count_line.n_atom_list = int(curr_list[2])
                    ctable.count_line.chiral_flag = int(curr_list[3])
                    ctable.count_line.n_stext_ents = int(curr_list[4])
                    ctable.count_line.n_add_props = int(curr_list[5])
                    ctable.count_line.version = curr_list[6]
                except:
                    raise TypeError('Count line has wrong format. Count Line: ', row)
            else:
                curr_list = row.split()
                try:
                    if cont:
                        _ = int(curr_list[0])
                    else:
                        _ = int('')
                except ValueError:
                    if row.strip() != '':
                        if extraProp != '':
                            sdf_dict.extraInfo.append(extraProp + row)
                            extraProp = ''
                        elif curr_list[0].strip().lower() == 'm':
                            if curr_list[1].strip().lower() != 'end':
                                ctable.properties.append(row)
                            else:
                                cont = False
                                sdf_dict.header = head
                                sdf_dict.con_table = ctable
                        elif curr_list[0].strip() == '>':
                            if extraProp == '':
                                extraProp += row[1:][row[1:].find('<') + 1: row[1:].find('>')] + ': '

                        elif row[0] == '$':
                            records.append(sdf_dict)
                            self.files_.append(path)
                            sdf_dict: Sdf = Sdf(header=SdfHeader(title='', timestamp='', comment=''),
                                                con_table=SdfCTable(
                                                    count_line=SdfCountLine(n_atoms=0, n_bonds=0, n_atom_list=0,
                                                                            chiral_flag=0, n_stext_ents=0,
                                                                            n_add_props=0, version=''),
                                                    atoms=Atoms(elements=[], coordinates=[], extraInfo=[]),
                                                    bonds=[
                                                        SdfBond(first_atom=0, second_atom=0, bond_type=0, bond_stereo=0,
                                                                bond_topology=0, reacting_center_status=0)],
                                                    properties=[]),
                                                extraInfo=[])

                            head: SdfHeader = SdfHeader(title='', timestamp='', comment='')
                            ctable: SdfCTable = SdfCTable(
                                count_line=SdfCountLine(n_atoms=0, n_bonds=0, n_atom_list=0, chiral_flag=0,
                                                        n_stext_ents=0, n_add_props=0, version=''),
                                atoms=Atoms(elements=[], coordinates=[], extraInfo=[]),
                                bonds=[SdfBond(first_atom=0, second_atom=0, bond_type=0, bond_stereo=0, bond_topology=0,
                                               reacting_center_status=0)],
                                properties=[]
                            )
                            line_cnt = -1
                        else:
                            try:
                                ctable.atoms.elements.append(_clean_key(curr_list[3]))
                                ctable.atoms.coordinates.append(
                                    [float(curr_list[0]), float(curr_list[1]), float(curr_list[2])])
                                ctable.atoms.extraInfo.append(curr_list[4:])
                            except:
                                # empty row
                                pass
                except IndexError:
                    pass
                except Exception as e:
                    raise TypeError('File has malformed fields. Error: ', str(e))
                else:
                    try:
                        ctable.bonds.append(parse_obj_as(SdfBond, {
                            'first_atom': int(curr_list[0]),
                            'second_atom': int(curr_list[1]),
                            'bond_type': int(curr_list[2]),
                            'bond_stereo': int(curr_list[3]),
                            'bond_topology': int(curr_list[4]),
                            'reacting_center_status': int(curr_list[5]),

                        }))
                    except:
                        raise TypeError('Found bond with wrong format. Bond: ', row)

            line_cnt += 1

        if len(records) > 1:
            return records

        try:
            return records[0]
        except:
            return sdf_dict

    def _parse_dataframe(self, file: Sdf, filename: str) -> List[pd.DataFrame]:
        """
        Parse mol or sdf files in a pandas dataframe.

         Parameters
        ----------
        file: jaqpotpy.models.material_models.Sdf
            The Sdf structrure of a parsed file

        filename: str
            The name of the parsed file

        Returns
        -------
        List[pd.DataFrame()]
        """
        atoms = pd.DataFrame()
        bonds = pd.DataFrame()

        for i in range(len(file.con_table.atoms.elements)):
            d = {}
            d['file'] = filename
            d['element'] = file.con_table.atoms.elements[i]
            d['x'] = file.con_table.atoms.coordinates[i][0]
            d['y'] = file.con_table.atoms.coordinates[i][1]
            d['z'] = file.con_table.atoms.coordinates[i][2]

            atoms = pd.concat([atoms, pd.DataFrame(d, index=[0])]).reset_index(drop=True)

        for bond in file.con_table.bonds:
            d = bond.dict()
            d['file'] = filename
            bonds = pd.concat([bonds, pd.DataFrame(d, index=[0])]).reset_index(drop=True)

        return [atoms, bonds]



from jaqpotpy.datasets.dataset_base import MaterialDataset
try:
    from pymatgen.core.composition import Composition
    from pymatgen.core.structure import Lattice, Structure
    from pymatgen.util.typing import CompositionLike
except ModuleNotFoundError:
    Lattice = Structure = CompositionLike = Composition = None

    pass
from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
# from jaqpotpy.descriptors import MolGraphConvFeaturizer, TorchMolGraphConvFeaturizer
from typing import Iterable, Any, Union, Dict
import pandas as pd
import os
import inspect
import numpy as np
import pickle
from torch.utils.data import Dataset


class CompositionDataset(MaterialDataset):
    """
    Reads CSV with compositions, experimental features and endpoints.

    Note
    ----
    Compositions may also be passed as an Iterable of stings representations, if no CSV file is available.
    In this case the endpoints should be passed separetely.
    """
    def __init__(self,  path: str = None, compositions: Union[str, Iterable[CompositionLike]] = None, y_cols: Union[str, Iterable[str]] = None,
                 keep_cols: Union[str, Iterable[str]]=None, y: Iterable[Any] = None, featurizer: MaterialFeaturizer = None, task: str = "regression") -> None:
        """
        A Composition dataset is a Tabular dataset and is defined either by an Iterable of compositions
        or by a CSV. In addition, the endpoints of the dataset should be defined.

        Parameters
        ----------
        path: str
            A path to a CSV file containing the compositions and the endpoints. The compositions in the
            CSV should be in a string format (e.g. "FeO", "Fe3O2", etc.). The file will be passed as a pandas
            DataFrame and descriptors will be the featurization of the compositions in the CSV file.
            If a path is passed, the, the compositions parameter should be filled with the column name
            of the compositions and the y_cols parameter should be filled with the column name of the
            endpoints. In addition, the y parameter may be omitted. Finally, please note that the keep_cols
            parameter can be used to keep and use some other columns from the dataset as well as the
            created descriptors.

        compositions: str or an iterable of pymatgen.util.typing.CompositionLike
            If the path parameter was passed, the compositions parameter should contain the column name of
            the compositions. If the path parameter was not passed then the compositions parameter is mandatory
            and should be an Iterable of pymatgen.util.typing.CompositionLike from which the featurizer
            shall create the descriptors.

        y_cols: str or an iterable of str
            This parameter is taken into account in the case of passing a CSV file. Use the y_cols to
            define the columns in the CSV that contain the endpoints for the modeling. Please note
            that defining the enpoints from the CSV file and not externally isn't mandatory,
            although it is highly recommended.

        keep_cols: str or an iterable of str
            This parameter is taken into account in the case of passing a CSV file. Use the y_cols to
            define the columns in the CSV that should be kept along with the other features that will
            be created from the featurizer.

        y: Iterable
            The y parameter is mandatory in the case where a CSV file was NOT provided. With the y
            parameter the modellers should pass the endpoint data. If a CSV file was passed,
            it is highly recommended to use the y_cols parameter to define the endpoints.

        featurizer: jaqpotpy.descriptors.MaterialFeaturizer
            The featurizer that will take the compositions and create the features for the modelling.

        task: str
            The task of the dataset. Either "regression" (default) or "classification".
        """
        super().__init__(path=path, x_cols=keep_cols, y_cols=y_cols)
        self._y = y
        self._x = None
        self._df = None
        self._x_cols_all = None
        if path:
            self.materials_col = compositions
            self.materials = None
        else:
            self.materials_col = None
            self.materials = compositions
        self._task = task
        self.featurizer: MaterialFeaturizer = featurizer
        self.indices: [] = None



        # self.create()

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdb", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpotpy_dataset" + ".jdb", 'wb') as f:
                pickle.dump(self, f)

    def create(self):

        if self.path:
            data = pd.read_csv(self.path)
            self.materials = data[self.materials_col].to_list()
            y = data[self.y_cols]
            if self.x_cols:
                xs = data[self.x_cols].copy()
            else:
                xs = data.drop(self.y_cols, axis=1).copy()
        else:
            xs = pd.DataFrame()

        if isinstance(self.materials[0], str):
            self._materials_strings = self.materials
        else:
            try:
                self._materials_strings = [Composition(item).formula for item in self.materials]
            except:
                self._materials_strings = []
                raise Warning('Found bad format in a Composition')

        self.indices = list(range(len(self._materials_strings)))
        descriptors = self.featurizer.featurize_dataframe(self.materials)
        if self._y:
            if self.y_cols:
                raise ValueError('Both y and y_cols parameters were passed. Please provide only \
                one of them. The y_cols is strongly recommended')
            else:
                self.y_cols = 'Endpoint'
                y = pd.DataFrame(data=self._y, columns=[self.y_cols])
        else:
            self.y_cols = list(y)
            self.y = list(y)

        if self.x_cols:
            self.df = pd.concat([descriptors, xs, y], axis=1)
            self._x_cols_all = list(descriptors) + list(xs)
        else:
            self.df = pd.concat([descriptors, y], axis=1)
            self._x_cols_all = list(descriptors)

        if self.X is None:
            if self.x_cols:
                xs = data[self.x_cols]
                if not xs.empty:
                    self._x = list(descriptors) + list(xs)
                else:
                    self._x = list(descriptors)
            else:
                self._x = list(descriptors)
        return self

    def __get_X__(self):
        return self.df[self.X].to_numpy()

    def __get_Y__(self):
        # return self.df[self.y].to_numpy()
        return self.y

    def __get__(self):
        return self.df

    def __getitem__(self, idx):
        # print(self.df[self.X].iloc[idx].values)
        # print(type(self.df[self.X].iloc[idx].values))
        X = self.df[self.X].iloc[idx].values
        y = self.df[self.y].iloc[idx].to_numpy()
        return X, y

    def __len__(self):
        if self.df is None:
            self.create()
        return len(self.df)

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
        #   # for list
        return self.__class__.__name__


class StructureDataset(MaterialDataset):
    """
    Creates a dataset for given structures. Structures may be passed as file paths for
    material files (e.g. cif, extxyz etc.) or in terms of any type capable of creating
    a pymatgen.core.Structure object.

    Note
    ----
    Compositions may also be passed as an Iterable of stings representations, if no CSV file is available.
    In this case the endpoints should be passed separetely.
    """
    def __init__(self,  path: Iterable[str] = None, structures: Union[Iterable[Structure], Iterable[Dict]] = None, ext_dataset: Union[pd.DataFrame, str] = None,
                 y_cols: Union[str, Iterable[str]] = None, keep_cols: Union[str, Iterable[str]]=None, y: Iterable[Any] = None,
                 featurizer: MaterialFeaturizer = None, task: str = "regression") -> None:
        """
        A Structure dataset is defined either by an Iterable of material files (containing information on the structure)
        or by an Interable of Structures. In addition, the endpoints of the dataset should be defined.

        Parameters
        ----------
        path: Iterable of str
            An Iterable of material file paths containing information on the structures of the materials
            from which the pymatgen.util.structure.Structure objects will be created. The featurizer shall
            create the descriptors from the structures. If a path is passed, the structures parameter should
            be omitted.

        structures: Iterable of pymatgen.util.structure.Structure or Iterable of dictionaries.
            If the path parameter was passed, the structures parameter should be omitted.
            If the path parameter was not passed then the structures parameter is mandatory
            and should be either an Iterable of pymatgen.util.structure.Structure or an Iterable of dict
            from which pymatgen.util.structure.Structure objects will be created.
            The featurizer shall create the descriptors from the structures.

        y_cols: str or an iterable of str
            This parameter is taken into account in the case of passing a CSV file. Use the y_cols to
            define the columns in the CSV that contain the endpoints for the modeling. Please note
            that defining the enpoints from the CSV file and not externally isn't mandatory,
            although it is highly recommended.

        keep_cols: str or an iterable of str
            This parameter is taken into account in the case of passing a CSV file. Use the y_cols to
            define the columns in the CSV that should be kept along with the other features that will
            be created from the featurizer.

        y: Iterable
            The y parameter is mandatory in the case where a CSV file was NOT provided. With the y
            parameter the modellers should pass the endpoint data. If a CSV file was passed,
            it is highly recommended to use the y_cols parameter to define the endpoints.

        featurizer: jaqpotpy.descriptors.MaterialFeaturizer
            The featurizer that will take the compositions and create the features for the modelling.

        task: str
            The task of the dataset. Either "regression" (default) or "classification".
        """
        super().__init__(path=path, x_cols=keep_cols, y_cols=y_cols)
        self._y = y
        self._x = None
        self._df = None
        self._external = ext_dataset
        self._x_cols_all = None
        self.materials_col = None
        self.materials = structures
        self._task = task
        self.featurizer: MaterialFeaturizer = featurizer
        self.indices: [] = None

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdb", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpotpy_dataset" + ".jdb", 'wb') as f:
                pickle.dump(self, f)

    def create(self):

        xs = pd.DataFrame()

        if self._external:
            if isinstance(self._external, str):
                data = pd.read_csv(self._external)
            else:
                data = self._external.copy()

            if self.x_cols:
                xs = data[self.x_cols].copy()
            else:
                xs = data.drop(self.y_cols, axis=1).copy()

        if self.path:
            structures = [Structure.from_file(item) for item in self.path]
        else:
            if isinstance(self.materials[0], dict):
                try:
                    structures = [Structure.from_dict(item) for item in self.materials]
                except:
                    try:
                        structures = [Structure(item['lattice'], item['species'], item['coords']) for item in self.materials]
                    except:
                        raise ValueError("The sturctures parameter should be either a dictinary view of a pymatgen.core.Structure or a dictionary \
                        with the keys 'lattice', 'species' and 'coords' and the appropriate values.")
            else:
                structures = self.materials

        self.materials = structures
        self._materials_strings = [item.composition.reduced_composition.formula for item in structures]
        self.indices = list(range(len(self._materials_strings)))
        descriptors = self.featurizer.featurize_dataframe(self.materials)

        if not self.y_cols:
            self.y_cols = 'Endpoint'

        y = pd.DataFrame(data=self._y, columns=[self.y_cols])

        self.y = list(y)

        if self.x_cols:
            self.df = pd.concat([descriptors, xs, y], axis=1)
            self._x_cols_all = list(descriptors) + list(xs)
        else:
            self.df = pd.concat([descriptors, y], axis=1)
            self._x_cols_all = list(descriptors)

        if self.X is None:
            if self.x_cols:
                xs = data[self.x_cols]
                if not xs.empty:
                    self._x = list(descriptors) + list(xs)
                else:
                    self._x = list(descriptors)
            else:
                self._x = list(descriptors)
        return self

    def __get_X__(self):
        return self.df[self.X].to_numpy()

    def __get_Y__(self):
        return self.df[self.y].to_numpy()

    def __get__(self):
        return self.df

    def __getitem__(self, idx):
        # print(self.df[self.X].iloc[idx].values)
        # print(type(self.df[self.X].iloc[idx].values))
        X = self.df[self.X].iloc[idx].values
        y = self.df[self.y].iloc[idx].to_numpy()
        return X, y

    def __len__(self):
        if self.df is None:
            self.create()
        return len(self.df)

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
        #   # for list
        return self.__class__.__name__